import unittest

import numpy as np
from CADETProcess import CADETProcessError
from CADETProcess.processModel import ComponentSystem


class TestComponents(unittest.TestCase):
    def setUp(self):
        self.component_system_0 = ComponentSystem(2)

        self.component_system_1 = ComponentSystem(["A", "B"])

        self.component_system_2 = ComponentSystem()
        self.component_system_2.add_component("A")
        self.component_system_2.add_component("B", species=["B+", "B-"])

        self.component_system_3 = ComponentSystem()
        self.component_system_3.add_component(
            "Ammonia",
            species=["NH4+", "NH3"],
            charge=[1, 0],
        )
        self.component_system_3.add_component(
            "Lysine", ["Lys2+", "Lys+", "Lys", "Lys"], [2, 1, 0, -1]
        )
        self.component_system_3.add_component(
            "H+",
            charge=1,
        )

        self.component_system_4 = ComponentSystem(2)
        self.component_system_4.add_component("manual_label")

        self.component_system_5 = ComponentSystem()
        self.component_system_5.add_component(
            "A", species=["A+", "A-"], molecular_weight=[1, 0]
        )

        self.component_system_6 = ComponentSystem()
        self.component_system_6.add_component("A", species=["A+", "A-"], density=[1, 0])

    def test_names(self):
        names_expected = ["0", "1"]
        names = self.component_system_0.names
        np.testing.assert_equal(names, names_expected)

        names_expected = ["A", "B"]
        names = self.component_system_1.names
        np.testing.assert_equal(names, names_expected)

        names_expected = ["A", "B"]
        names = self.component_system_2.names
        np.testing.assert_equal(names, names_expected)

        names_expected = ["Ammonia", "Lysine", "H+"]
        names = self.component_system_3.names
        np.testing.assert_equal(names, names_expected)

        names_expected = ["0", "1", "manual_label"]
        names = self.component_system_4.names
        np.testing.assert_equal(names, names_expected)

    def test_duplicate_name(self):
        with self.assertRaises(CADETProcessError):
            self.component_system_1.add_component("A")

    def test_species(self):
        species_expected = ["0", "1"]
        species = self.component_system_0.species
        np.testing.assert_equal(species, species_expected)

        species_expected = ["A", "B"]
        species = self.component_system_1.species
        np.testing.assert_equal(species, species_expected)

        species_expected = ["A", "B+", "B-"]
        species = self.component_system_2.species
        np.testing.assert_equal(species, species_expected)

        species_expected = ["NH4+", "NH3", "Lys2+", "Lys+", "Lys", "Lys", "H+"]
        species = self.component_system_3.species
        np.testing.assert_equal(species, species_expected)

        species_expected = ["0", "1", "manual_label"]
        species = self.component_system_4.species
        np.testing.assert_equal(species, species_expected)

    def test_indices(self):
        indices_expected = {
            "Ammonia": [0, 1],
            "Lysine": [2, 3, 4, 5],
            "H+": [6],
        }
        indices = self.component_system_3.indices
        np.testing.assert_equal(indices, indices_expected)

    def test_n_comp(self):
        n_components_expected = 2
        n_components = self.component_system_2.n_components
        self.assertEqual(n_components_expected, n_components)

        n_comp_expected = 3
        n_comp = self.component_system_2.n_comp
        self.assertEqual(n_comp_expected, n_comp)

        n_components_expected = 3
        n_components = self.component_system_3.n_components
        self.assertEqual(n_components_expected, n_components)

        n_comp_expected = 7
        n_comp = self.component_system_3.n_comp
        self.assertEqual(n_comp_expected, n_comp)

    def test_charge(self):
        charges_expected = [1, 0, 2, 1, 0, -1, 1]
        charges = self.component_system_3.charges
        np.testing.assert_equal(charges_expected, charges)

    def test_molecular_weights(self):
        molecular_weights_expected = [1, 0]
        molecular_weights = self.component_system_5.molecular_weights
        np.testing.assert_equal(molecular_weights_expected, molecular_weights)

    def test_densities(self):
        densities_expected = [1, 0]
        densities = self.component_system_6.densities
        np.testing.assert_equal(densities_expected, densities)


if __name__ == "__main__":
    unittest.main()



import functools
import warnings
from typing import Any, Callable, Dict

__all__ = ["deprecated_alias", "rename_kwargs"]


def deprecated_alias(**aliases: str) -> Callable:
    """
    Add alias for deprecated function arguments.

    Parameters
    ----------
    **aliases : str
        Keyword arguments where keys are old parameter names and values are new parameter names

    Returns
    -------
    Callable
        A decorator function that wraps the original function

    Examples
    --------
    @deprecated_alias(old_name='new_name')
    def example_function(new_name):
         return new_name
    """

    # Decorator function: takes the f and returns a f with the new argument names
    def decorator(f: Callable) -> Callable:
        @functools.wraps(f)
        def wrap_decorated_argument(*args: Any, **kwargs: Any) -> Any:
            rename_kwargs(f.__name__, kwargs, aliases)
            return f(*args, **kwargs)

        return wrap_decorated_argument

    return decorator


def rename_kwargs(
    func_name: str, kwargs: Dict[str, Any], aliases: Dict[str, str]
) -> None:
    """
    Automatically rename deprecated function arguments.

    Parameters
    ----------
    func_name : str
        Name of the function being decorated
    kwargs : Dict[str, Any]
        Dictionary of keyword arguments passed to the function
    aliases : Dict[str, str]
        Dictionary mapping old parameter names to new ones

    Returns
    -------
    None

    Raises
    ------
    TypeError
        If both old and new parameter names are used simultaneously

    Examples
    --------
    rename_kwargs('example_function', {'old_name': 'value'}, {'old_name': 'new_name'})
    """
    for alias, new in aliases.items():
        if alias in kwargs:
            if new in kwargs:
                raise TypeError(
                    f"{func_name} received both {alias} and {new} as arguments!"
                    f" {alias} is deprecated, use {new} instead."
                )
            warnings.warn(
                message=(
                    f"`{alias}` is deprecated as an argument to `{func_name}`; use `{new}` instead."
                ),
                category=DeprecationWarning,
                stacklevel=2,
            )
            kwargs[new] = kwargs.pop(alias)



"""
=========================================
Reference (:mod:`CADETProcess.reference`)
=========================================

.. currentmodule:: CADETProcess.reference

This module provides functionality for setting up reference solutions used for
comparison with ``SimulationResults``

.. autosummary::
    :toctree: generated/

    ReferenceBase
    ReferenceIO

"""  # noqa

from typing import Any, Optional

import numpy as np
import numpy.typing as npt

from CADETProcess import CADETProcessError
from CADETProcess.processModel import ComponentSystem
from CADETProcess.solution import SolutionBase, SolutionIO

__all__ = ["ReferenceBase", "ReferenceIO", "FractionationReference"]


class ReferenceBase(SolutionBase):
    """
    Class representing references to be compared with SimulationResults.

    See Also
    --------
    CADETProcess.solution.SolutionBase
    """

    pass


class ReferenceIO(ReferenceBase, SolutionIO):
    """
    A class representing reference data of inlet or outlet concentration profiles.

    Attributes
    ----------
    name : str
        The name of the reference.
    component_system : ComponentSystem
        The reference component system.
    time : np.ndarray
        The time points for the reference.
    solution : np.ndarray
        The reference solution values.
    flow_rate : np.ndarray
        The flow rates for the reference.

    See Also
    --------
    CADETProcess.reference.ReferenceBase
    CADETProcess.solution.SolutionIO
    """

    def __init__(
        self,
        name: str,
        time: npt.ArrayLike,
        solution: npt.ArrayLike,
        flow_rate: Optional[float | npt.ArrayLike] = None,
        component_system: Optional[ComponentSystem] = None,
    ) -> None:
        """
        Initialize a ReferenceIO object.

        Parameters
        ----------
        name : str
            The name of the reference.
        time : array-like
            The time points for the reference.
        solution : array-like
            The reference solution values with shape = (n_time, n_comp).
        flow_rate : array-like or float, optional
            The flow rates for the reference.
            If not provided, flow rate of 1 is assumed.
        component_system : ComponentSystem, optional
            The reference component system.
            If not provided, a ComponentSystem with the same number of components as the
            solution is created.

        Raises
        ------
        TypeError
            If the provided time, solution, or flow rate are not array-like.
        ValueError
            If the time and solution arrays are not the same length.
            If the flow rate array and time array are not the same length.
        """
        time = np.array(time, dtype=np.float64).reshape(-1)

        if solution.shape[0] != len(time):
            raise ValueError(
                "Solution had the wrong shape. Solution needs the shape (time, n_comp)."
            )

        solution = np.array(solution, ndmin=2, dtype=np.float64).reshape(len(time), -1)

        if component_system is None:
            n_comp = solution.shape[1]
            component_system = ComponentSystem(n_comp)

        if flow_rate is None:
            flow_rate = 1
        if isinstance(flow_rate, (int, float)):
            flow_rate = flow_rate * np.ones(time.shape)

        super().__init__(name, component_system, time, solution, flow_rate)


class FractionationReference(ReferenceBase):
    """
    A class representing reference data of fractionation data.

    Attributes
    ----------
    name : str
        The name of the reference.
    component_system : ComponentSystem
        The reference component system.
    time : np.ndarray
        The time points for the reference.
    solution : np.ndarray
        The reference solution values.

    See Also
    --------
    CADETProcess.reference.ReferenceBase
    CADETProcess.fractionation.Fraction
    """

    dimensions = SolutionBase.dimensions + ["component_coordinates"]

    def __init__(
        self,
        name: str,
        fractions: list,
        component_system: Optional[ComponentSystem] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize FractionationReference object."""
        from CADETProcess.fractionation import Fraction

        for frac in fractions:
            if not isinstance(frac, Fraction):
                raise TypeError("Expected Fraction.")
            if frac.start is None or frac.end is None:
                raise CADETProcessError("Fractionation times must be provided.")
            if not frac.end > frac.start:
                raise CADETProcessError("Fraction end time must be greater than start.")

        self.fractions = fractions

        time = np.array([(frac.start + frac.end) / 2 for frac in self.fractions])
        solution = np.array([frac.mass / frac.volume for frac in self.fractions])

        if component_system is None:
            n_comp = solution.shape[1]
            component_system = ComponentSystem(n_comp)

        super().__init__(name, component_system, time, solution)



"""
CADET-Process
========

CADET-Process is a Python package for modelling, simulating and optimizing
advanced chromatographic systems. It serves as an inteface for CADET, but also
for other solvers.

See https://cadet-process.readthedocs.io for complete documentation.
"""

# Version information
name = "CADET-Process"
__version__ = "0.11.1"

# Imports
from .CADETProcessError import *

from . import log

from .settings import Settings

settings = Settings()

from . import sysinfo
from . import numerics
from . import dataStructure
from . import transform
from . import plotting
from . import dynamicEvents
from . import processModel
from . import smoothing
from . import solution
from . import reference
from .simulationResults import SimulationResults
from . import metric
from . import performance
from . import optimization
from . import comparison
from . import stationarity
from . import simulator
from . import fractionation
from . import equilibria
from . import modelBuilder
from . import tools



import shutil
import time
import unittest
from abc import ABC, abstractmethod

import numpy as np
from addict import Dict
from CADETProcess import settings
from CADETProcess.dataStructure import (
    Float,
    List,
    NdPolynomial,
    Polynomial,
    SizedList,
    SizedNdArray,
    Structure,
)
from CADETProcess.optimization import OptimizationProblem

from tests.optimization_problem_fixtures import (
    LinearConstraintsSooTestProblem2,
    LinearEqualityConstraintsSooTestProblem,
)


class EvaluationObject(Structure):
    uninitialized = None
    scalar_param = Float(default=1)
    scalar_param_2 = Float(default=1)
    list_param = List()
    sized_list_param = SizedList(size=2, default=[1, 2])
    sized_list_param_single = SizedList(size=1, default=1)
    sized_list_param_no_default = SizedList(size=2)
    nd_array = SizedNdArray(size=(2, 2))
    polynomial_param = Polynomial(n_coeff=2, default=0)
    polynomial_param_no_default = Polynomial(n_coeff=2)
    nd_polynomial_param = NdPolynomial(size=(2, 4), default=0)

    _parameters = [
        "uninitialized",
        "scalar_param",
        "scalar_param_2",
        "list_param",
        "sized_list_param",
        "sized_list_param_single",
        "sized_list_param_no_default",
        "nd_array",
        "polynomial_param",
        "polynomial_param_no_default",
        "nd_polynomial_param",
    ]

    def __init__(self, name="Dummy"):
        self.name = name
        super().__init__()

    def __str__(self):
        return self.name


class Evaluator(ABC):
    @abstractmethod
    def evaluate(self, request):
        pass

    def __call__(self, *args, **kwargs):
        return self.evaluate(*args, **kwargs)

    def __str__(self):
        return self.__class__.__name__


class ExpensiveEvaluator(Evaluator):
    """Mockup for an expensive evaluator"""

    def evaluate(self, evaluation_object):
        print("Run expensive evaluator; this takes forever...")

        expensive_results = Dict()
        expensive_results.result = np.array(2 * [evaluation_object.scalar_param])

        time.sleep(1)

        return expensive_results


class CheapEvaluator(Evaluator):
    """Mockup for a cheap evaluator"""

    def evaluate(self, expensive_results):
        print("Run cheap evaluator.")
        cheap_results = Dict()
        cheap_results.result_1 = expensive_results.result * 2
        cheap_results.result_2 = expensive_results.result * 2

        return cheap_results


def sum_x(x):
    print("run sum(x)")
    return [sum(x)]


def min_results_1(cheap_results):
    print("run MinResult1")
    score = cheap_results.result_1 * 2
    return score.tolist()


def min_results_2(cheap_results):
    print("run MinResult2")
    score = cheap_results.result_2 * 3
    return score.tolist()


from CADETProcess import CADETProcessError
from CADETProcess.optimization import OptimizationVariable


class Test_OptimizationVariable(unittest.TestCase):
    def setUp(self):
        self.evaluation_object = EvaluationObject()

    def test_index(self):
        var = OptimizationVariable(
            "no_index_required",
            evaluation_objects=[self.evaluation_object],
            parameter_path="scalar_param",
        )
        var.value = 1
        np.testing.assert_equal(var.value, 1)
        np.testing.assert_equal(self.evaluation_object.scalar_param, 1)

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "parameter_not_sized",
                evaluation_objects=[self.evaluation_object],
                parameter_path="scalar_param",
                indices=1,
            )

        var = OptimizationVariable(
            "list_index",
            evaluation_objects=[self.evaluation_object],
            parameter_path="sized_list_param",
            indices=1,
        )
        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(var.indices, [[(1,)]])
        np.testing.assert_equal(var.full_indices, [[(1,)]])
        np.testing.assert_equal(self.evaluation_object.sized_list_param, [1, 2])

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "index_exceeds_list_size",
                evaluation_objects=[self.evaluation_object],
                parameter_path="sized_list_param",
                indices=2,
            )

        # Test uninitialized value
        with self.assertRaises(CADETProcessError):
            var = OptimizationVariable(
                "uninitialized_attribute",
                evaluation_objects=[self.evaluation_object],
                parameter_path="uninitialized",
                indices=2,
            )

    def test_significant_digits(self):
        var_no_precision = OptimizationVariable(
            "not_rounded_variable",
            evaluation_objects=[self.evaluation_object],
            parameter_path="scalar_param",
            significant_digits=None,
        )
        var_no_precision.value = 1.1234
        np.testing.assert_equal(var_no_precision.value, 1.1234)
        np.testing.assert_equal(self.evaluation_object.scalar_param, 1.1234)

        var_with_precision = OptimizationVariable(
            "rounded_variable",
            evaluation_objects=[self.evaluation_object],
            parameter_path="scalar_param",
            significant_digits=2,
        )
        var_with_precision.value = 1.1234
        np.testing.assert_equal(var_with_precision.value, 1.1)
        np.testing.assert_equal(self.evaluation_object.scalar_param, 1.1)

    def test_sized_list_single(self):
        """
        Extra test for array with single entry.

        See also: https://github.com/fau-advanced-separations/CADET-Process/pull/68
        """
        # No indices
        var = OptimizationVariable(
            "sized_list_param_single",
            evaluation_objects=[self.evaluation_object],
            parameter_path="sized_list_param_single",
        )
        np.testing.assert_equal(var.indices, [[(slice(None, None, None),)]])
        var.value = 1
        np.testing.assert_equal(var.value, 1)
        np.testing.assert_equal(self.evaluation_object.sized_list_param_single, [1])

        # Explicit index
        var = OptimizationVariable(
            "sized_list_param_single",
            evaluation_objects=[self.evaluation_object],
            parameter_path="sized_list_param_single",
            indices=0,
        )
        np.testing.assert_equal(var.indices, [[(0,)]])
        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(self.evaluation_object.sized_list_param_single, [2])

        # Raise Exception for exceeding index
        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "sized_list_param_single",
                evaluation_objects=[self.evaluation_object],
                parameter_path="sized_list_param_single",
                indices=1,
            )

    def test_nd_array(self):
        var = OptimizationVariable(
            "nd_array_index",
            evaluation_objects=[self.evaluation_object],
            parameter_path="nd_array",
            indices=(0, 0),
        )
        var.value = 1
        np.testing.assert_equal(var.value, 1)
        np.testing.assert_equal(var.indices, [[(0, 0)]])
        np.testing.assert_equal(var.full_indices, [[(0, 0)]])
        np.testing.assert_equal(
            self.evaluation_object.nd_array, [[1, np.nan], [np.nan, np.nan]]
        )

        var = OptimizationVariable(
            "nd_array_slice",
            evaluation_objects=[self.evaluation_object],
            parameter_path="nd_array",
            indices=np.s_[:, 0],
        )
        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(var.indices, [[(slice(None, None, None), 0)]])
        np.testing.assert_equal(var.full_indices, [[(0, 0), (1, 0)]])
        np.testing.assert_equal(
            self.evaluation_object.nd_array, [[2, np.nan], [2, np.nan]]
        )

        var = OptimizationVariable(
            "nd_array_all",
            evaluation_objects=[self.evaluation_object],
            parameter_path="nd_array",
        )
        var.value = 3
        np.testing.assert_equal(var.value, 3)
        np.testing.assert_equal(var.indices, [[(slice(None, None, None),)]])
        np.testing.assert_equal(var.full_indices, [[(0, 0), (0, 1), (1, 0), (1, 1)]])
        np.testing.assert_equal(self.evaluation_object.nd_array, [[3, 3], [3, 3]])

    def test_polynomial(self):
        var = OptimizationVariable(
            "polynomial",
            evaluation_objects=self.evaluation_object,
            parameter_path="polynomial_param",
            indices=0,
        )

        var.value = 1
        np.testing.assert_equal(var.value, 1)
        np.testing.assert_equal(var.indices, [[(0,)]])
        np.testing.assert_equal(var.full_indices, [[(0,)]])
        np.testing.assert_equal(self.evaluation_object.scalar_param, 1)

        # Check convenience function; No index should modify constant coefficient and
        # and set rest to 0
        var = OptimizationVariable(
            "polynomial_convenience_1D",
            evaluation_objects=self.evaluation_object,
            parameter_path="polynomial_param",
        )
        var.value = 1
        np.testing.assert_equal(var.value, 1)
        np.testing.assert_equal(self.evaluation_object.polynomial_param, [1, 0])

        var = OptimizationVariable(
            "polynomial_convenience_ND_all",
            evaluation_objects=self.evaluation_object,
            parameter_path="nd_polynomial_param",
        )
        var.value = 1
        np.testing.assert_equal(var.value, 1)
        np.testing.assert_equal(var.indices, [[(slice(None, None, None),)]])
        np.testing.assert_equal(
            var.full_indices,
            [[(0, 0), (0, 1), (0, 2), (0, 3), (1, 0), (1, 1), (1, 2), (1, 3)]],
        )
        np.testing.assert_equal(
            self.evaluation_object.nd_polynomial_param, [[1, 0, 0, 0], [1, 0, 0, 0]]
        )

        var = OptimizationVariable(
            "polynomial_convenience_ND_single",
            evaluation_objects=self.evaluation_object,
            parameter_path="nd_polynomial_param",
            indices=0,
        )
        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(var.indices, [[(0,)]])
        np.testing.assert_equal(var.full_indices, [[(0, 0), (0, 1), (0, 2), (0, 3)]])
        np.testing.assert_equal(
            self.evaluation_object.nd_polynomial_param, [[2, 0, 0, 0], [1, 0, 0, 0]]
        )

        var = OptimizationVariable(
            "polynomial_convenience_ND_single",
            evaluation_objects=self.evaluation_object,
            parameter_path="nd_polynomial_param",
            indices=np.s_[0, :],
        )
        var.value = 3
        np.testing.assert_equal(var.value, 3)
        np.testing.assert_equal(var.indices, [[(0, slice(None, None, None))]])
        np.testing.assert_equal(var.full_indices, [[(0, 0), (0, 1), (0, 2), (0, 3)]])
        np.testing.assert_equal(
            self.evaluation_object.nd_polynomial_param, [[3, 3, 3, 3], [1, 0, 0, 0]]
        )

    def test_transform(self):
        var = OptimizationVariable(
            "scalar_param",
            evaluation_objects=[self.evaluation_object],
            parameter_path="scalar_param",
        )

        # Missing bounds
        with self.assertRaises(CADETProcessError):
            var = OptimizationVariable(
                "scalar_param",
                evaluation_objects=[self.evaluation_object],
                parameter_path="scalar_param",
                transform="auto",
            )


from tests.test_events import HandlerFixture


class Test_OptimizationVariableEvents(unittest.TestCase):
    def setUp(self):
        evaluation_object = HandlerFixture()

        evt = evaluation_object.add_event(
            "single_index_1D", "performer.array_1d", 1, indices=0, time=1
        )
        evt = evaluation_object.add_event(
            "single_index_ND", "performer.ndarray", 1, indices=(0, 0), time=2
        )

        evt = evaluation_object.add_event(
            "multi_index_1D", "performer.array_1d", [0, 1], indices=[0, 1], time=3
        )
        evt = evaluation_object.add_event(
            "multi_index_ND",
            "performer.ndarray",
            [0, 1],
            indices=[(0, 0), (0, 1)],
            time=4,
        )

        evt = evaluation_object.add_event(
            "nd_evt_state", "performer.ndarray", [[8, 7, 6, 5], [4, 3, 2, 1]], time=5
        )

        evt = evaluation_object.add_event(
            "array_1d_poly_full", "performer.array_1d_poly", 1, time=6
        )
        evt = evaluation_object.add_event(
            "array_1d_poly_indices",
            "performer.array_1d_poly",
            [1, 1],
            indices=[1, 2],
            time=7,
        )

        evt = evaluation_object.add_event(
            "ndarray_poly_full", "performer.ndarray_poly", 1, time=8
        )
        evt = evaluation_object.add_event(
            "ndarray_poly_indices",
            "performer.ndarray_poly",
            [1, 1],
            indices=[(0, 1), (1, -1)],
            time=9,
        )

        self.evaluation_object = evaluation_object

    def test_index(self):
        # Single Event.state entry
        var = OptimizationVariable(
            "single_index_1D",
            evaluation_objects=[self.evaluation_object],
            parameter_path="single_index_1D.state",
        )
        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(self.evaluation_object.single_index_1D.state, 2)
        np.testing.assert_equal(self.evaluation_object.performer.array_1d, [2, 1, 0, 0])

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "event_state_not_sized",
                evaluation_objects=[self.evaluation_object],
                parameter_path="single_index_1D.state",
                indices=1,
            )

        var = OptimizationVariable(
            "single_index_ND",
            evaluation_objects=[self.evaluation_object],
            parameter_path="single_index_ND.state",
        )
        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(self.evaluation_object.single_index_ND.state, 2)
        np.testing.assert_equal(
            self.evaluation_object.performer.ndarray, [[2, 7, 6, 5], [4, 3, 2, 1]]
        )

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "event_state_not_sized",
                evaluation_objects=[self.evaluation_object],
                parameter_path="single_index_ND.state",
                indices=1,
            )

        # 1D Event.state with multiple entries
        var = OptimizationVariable(
            "multi_index_1D",
            evaluation_objects=[self.evaluation_object],
            parameter_path="multi_index_1D.state",
            indices=0,
        )
        var.value = 3
        np.testing.assert_equal(var.value, 3)
        np.testing.assert_equal(self.evaluation_object.multi_index_1D.state, [3, 1])
        np.testing.assert_equal(self.evaluation_object.performer.array_1d, [3, 1, 0, 0])

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "index_exceeds_parameter",
                evaluation_objects=[self.evaluation_object],
                parameter_path="multi_index_1D.state",
                indices=2,
            )
        # ND Event.state with multiple entries (in 1D)
        var = OptimizationVariable(
            "multi_index_ND",
            evaluation_objects=[self.evaluation_object],
            parameter_path="multi_index_ND.state",
            indices=0,
        )
        var.value = 4
        np.testing.assert_equal(var.value, 4)
        np.testing.assert_equal(self.evaluation_object.multi_index_ND.state, [4, 1])
        np.testing.assert_equal(
            self.evaluation_object.performer.ndarray, [[4, 1, 6, 5], [4, 3, 2, 1]]
        )

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "index_exceeds_parameter",
                evaluation_objects=[self.evaluation_object],
                parameter_path="multi_index_ND.state",
                indices=2,
            )

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "index_exceeds_parameter_dimension",
                evaluation_objects=[self.evaluation_object],
                parameter_path="multi_index_ND.state",
                indices=(1, 1),
            )

        # ND Event.state with multiple entries in ND
        var = OptimizationVariable(
            "nd_evt_state",
            evaluation_objects=[self.evaluation_object],
            parameter_path="nd_evt_state.state",
            indices=(0, 0),
        )
        var.value = 5
        np.testing.assert_equal(var.value, 5)
        np.testing.assert_equal(
            self.evaluation_object.nd_evt_state.state, [[5, 7, 6, 5], [4, 3, 2, 1]]
        )
        np.testing.assert_equal(
            self.evaluation_object.performer.ndarray, [[5, 7, 6, 5], [4, 3, 2, 1]]
        )

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "index_exceeds_parameter",
                evaluation_objects=[self.evaluation_object],
                parameter_path="nd_evt_state.state",
                indices=(2, 2),
            )

    def test_polynomial(self):
        # Event state modifies entire 1D polyomial coefficients using `fill_values`
        var = OptimizationVariable(
            "array_1d_poly_full",
            evaluation_objects=self.evaluation_object,
            parameter_path="array_1d_poly_full.state",
        )

        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(self.evaluation_object.array_1d_poly_full.state, 2)
        np.testing.assert_equal(
            self.evaluation_object.performer.array_1d_poly, [2, 0, 0, 0]
        )

        # Event state modifies individual 1D polyomial coefficients
        var = OptimizationVariable(
            "array_1d_poly_indices",
            evaluation_objects=self.evaluation_object,
            parameter_path="array_1d_poly_indices.state",
            indices=0,
        )

        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(
            self.evaluation_object.array_1d_poly_indices.state, [2, 1]
        )
        np.testing.assert_equal(
            self.evaluation_object.performer.array_1d_poly, [2, 2, 1, 0]
        )

        # with self.assertRaises(ValueError):
        #     var = OptimizationVariable(
        #         'missing_index',
        #         evaluation_objects=self.evaluation_object,
        #         parameter_path='array_1d_poly_indices.state',
        #     )

        # Event state modifies entire ND polyomial coefficients using `fill_values`
        var = OptimizationVariable(
            "ndarray_poly_full",
            evaluation_objects=self.evaluation_object,
            parameter_path="ndarray_poly_full.state",
        )

        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(self.evaluation_object.ndarray_poly_full.state, 2)
        np.testing.assert_equal(
            self.evaluation_object.performer.ndarray_poly, [[2, 0, 0, 0], [2, 0, 0, 0]]
        )

        # Event state modifies individual 1D polyomial coefficients
        var = OptimizationVariable(
            "ndarray_poly_indices",
            evaluation_objects=self.evaluation_object,
            parameter_path="ndarray_poly_indices.state",
            indices=0,
        )

        var.value = 2
        np.testing.assert_equal(var.value, 2)
        np.testing.assert_equal(
            self.evaluation_object.ndarray_poly_indices.state, [2, 1]
        )
        np.testing.assert_equal(
            self.evaluation_object.performer.ndarray_poly, [[2, 2, 0, 0], [2, 0, 0, 1]]
        )

        # with self.assertRaises(ValueError):
        #     var = OptimizationVariable(
        #         'missing_index',
        #         evaluation_objects=self.evaluation_object,
        #         parameter_path='ndarray_poly_indices.state',
        #     )

    def test_multi_eval_obj(self):
        """Test setting indexed variables for multiple evaluation objects."""
        evaluation_object_2 = HandlerFixture()

        evt = evaluation_object_2.add_event(
            "multi_index_ND",
            "performer.ndarray",
            [0, 1, 2],
            indices=[(0, 0), (0, 1), (1, 1)],
            time=0,
        )

        var = OptimizationVariable(
            "multi_index_ND",
            evaluation_objects=[self.evaluation_object, evaluation_object_2],
            parameter_path="multi_index_ND.state",
            indices=0,
        )

        with self.assertRaises(IndexError):
            var = OptimizationVariable(
                "index_exceeds_one_eval_obj_state",
                evaluation_objects=[self.evaluation_object, evaluation_object_2],
                parameter_path="multi_index_ND.state",
                indices=2,
            )


def setup_dummy_eval_fun(n_metrics, rng=None):
    if rng is None:
        rng = np.random.default_rng(12345)

    def dummy_eval_fun(x):
        return rng.random(n_metrics)

    return dummy_eval_fun


def dummy_meta_score(f):
    return np.sum(f)


def setup_optimization_problem(
    n_vars=2,
    n_obj=1,
    n_lincon=0,
    n_lineqcon=0,
    n_nonlincon=0,
    n_meta=0,
    bounds=None,
    obj_fun=None,
    nonlincon_fun=None,
    lincons=None,
    lineqcons=None,
    use_diskcache=False,
):
    optimization_problem = OptimizationProblem("simple", use_diskcache=use_diskcache)

    for i_var in range(n_vars):
        if bounds is None:
            lb, ub = (0, 1)
        else:
            lb, ub = bounds[i_var]

        optimization_problem.add_variable(f"var_{i_var}", lb=lb, ub=ub)

    if n_lincon > 0:
        if lincons is None:
            lincons = [
                ([f"var_{i_var}", f"var_{i_var + 1}"], [1, -1], 0)
                for i_var in range(n_lincon)
            ]
        for opt_vars, lhs, b in lincons:
            optimization_problem.add_linear_constraint(opt_vars, lhs, b)

    if n_lineqcon > 0:
        if lineqcons is None:
            lineqcons = [
                ([f"var_{i_var}", f"var_{i_var + 1}"], [1, -1], 0)
                for i_var in range(n_lineqcon)
            ]
        for opt_vars, lhs, beq in lineqcons:
            optimization_problem.add_linear_equality_constraint(opt_vars, lhs, beq)

    if obj_fun is None:
        obj_fun = setup_dummy_eval_fun(n_obj)

    optimization_problem.add_objective(
        obj_fun, n_objectives=n_obj, labels=[f"f_{i}" for i in range(n_obj)]
    )

    if n_nonlincon > 0:
        if nonlincon_fun is None:
            nonlincon_fun = setup_dummy_eval_fun(n_nonlincon)
        optimization_problem.add_nonlinear_constraint(
            nonlincon_fun,
            n_nonlinear_constraints=n_nonlincon,
            labels=[f"g_{i}" for i in range(n_nonlincon)],
            bounds=0.5,
        )

    if n_meta > 0:
        optimization_problem.add_meta_score(dummy_meta_score)

    return optimization_problem


class Test_OptimizationProblemSimple(unittest.TestCase):
    def setUp(self):
        optimization_problem = OptimizationProblem("simple", use_diskcache=False)

        optimization_problem.add_variable("var_0", lb=0, ub=1)
        optimization_problem.add_variable("var_1", lb=0, ub=10)

        self.optimization_problem = optimization_problem

    def tearDown(self):
        shutil.rmtree("./results_simple", ignore_errors=True)
        settings.working_directory = None

    def test_variable_names(self):
        names_expected = ["var_0", "var_1"]
        names = self.optimization_problem.variable_names
        self.assertEqual(names_expected, names)

        # Variable already exists
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable("var_0")

    def test_bounds(self):
        lb_expected = [0, 0]
        lb = self.optimization_problem.lower_bounds
        np.testing.assert_almost_equal(lb_expected, lb)

        ub_expected = [1, 10]
        ub = self.optimization_problem.upper_bounds
        np.testing.assert_almost_equal(ub_expected, ub)

        # lb >= ub
        with self.assertRaises(ValueError):
            self.optimization_problem.add_variable("spam", lb=0, ub=0)


class Test_OptimizationProblemLinCon(unittest.TestCase):
    def setUp(self):
        optimization_problem = setup_optimization_problem(
            n_lincon=1, use_diskcache=False
        )

        self.optimization_problem = optimization_problem

    def test_add_linear_constraints(self):
        self.optimization_problem.add_linear_constraint("var_0")
        self.optimization_problem.add_linear_constraint(["var_0", "var_1"])

        self.optimization_problem.add_linear_constraint(["var_0", "var_1"], [2, 2])
        self.optimization_problem.add_linear_constraint(["var_0", "var_1"], [3, 3], 1)
        self.optimization_problem.add_linear_constraint(["var_0", "var_1"], 4)

        A_expected = np.array(
            [
                [1.0, -1.0],
                [1.0, 0.0],
                [1.0, 1.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [4.0, 4.0],
            ]
        )

        A = self.optimization_problem.A
        np.testing.assert_almost_equal(A, A_expected)

        b_expected = np.array([0, 0, 0, 0, 1, 0])
        b = self.optimization_problem.b
        np.testing.assert_almost_equal(b, b_expected)

        # Variable does not exist
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_linear_constraint("inexistent")

        # Incorrect shape
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_linear_constraint("var_0", [])

    def test_add_linear_equality_constraints(self):
        self.optimization_problem.add_linear_equality_constraint("var_0")
        self.optimization_problem.add_linear_equality_constraint(["var_0", "var_1"])

        self.optimization_problem.add_linear_equality_constraint(
            ["var_0", "var_1"], [2, 2]
        )
        self.optimization_problem.add_linear_equality_constraint(
            ["var_0", "var_1"], [3, 3], 1
        )
        self.optimization_problem.add_linear_equality_constraint(["var_0", "var_1"], 4)

        Aeq_expected = np.array(
            [
                [1.0, 0.0],
                [1.0, 1.0],
                [2.0, 2.0],
                [3.0, 3.0],
                [4.0, 4.0],
            ]
        )

        Aeq = self.optimization_problem.Aeq
        np.testing.assert_almost_equal(Aeq, Aeq_expected)

        beq_expected = np.array([0, 0, 0, 1, 0])
        beq = self.optimization_problem.beq
        np.testing.assert_almost_equal(beq, beq_expected)

        # Variable does not exist
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_linear_equality_constraint("inexistent")

        # Incorrect shape
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_linear_equality_constraint("var_0", [])

    def test_initial_values(self):
        x0_chebyshev_expected = [0.29289322, 0.70710678]
        x0_chebyshev = self.optimization_problem.get_chebyshev_center(
            include_dependent_variables=True
        )
        np.testing.assert_almost_equal(x0_chebyshev, x0_chebyshev_expected)

        x0_seed_1_expected = [[0.3448127, 0.9138096]]
        x0_seed_1 = self.optimization_problem.create_initial_values(1, seed=1)
        np.testing.assert_almost_equal(x0_seed_1, x0_seed_1_expected)

        x0_seed_1_random = self.optimization_problem.create_initial_values(1)

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_1_random, x0_seed_1_expected)

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_1_random, x0_chebyshev_expected)

        x0_seed_10_expected = [
            [0.34481272, 0.91380963],
            [0.13935918, 0.7377129],
            [0.5219434, 0.63404393],
            [0.23933603, 0.8127269],
            [0.30483107, 0.980534],
            [0.11605111, 0.86206301],
            [0.1869498, 0.99896377],
            [0.55986142, 0.65237364],
            [0.49963092, 0.98791549],
            [0.6839696, 0.92668444],
        ]
        x0_seed_10 = self.optimization_problem.create_initial_values(10, seed=1)
        np.testing.assert_almost_equal(x0_seed_10, x0_seed_10_expected)

        x0_seed_10_random = self.optimization_problem.create_initial_values(10)

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_10_random, x0_seed_10_expected)


class Test_OptimizationProblemDepVar(unittest.TestCase):
    def setUp(self):
        optimization_problem = OptimizationProblem("simple", use_diskcache=False)

        optimization_problem.add_variable("foo", lb=0, ub=1)
        optimization_problem.add_variable("bar", lb=0, ub=1)
        optimization_problem.add_variable("spam", lb=0, ub=1)
        optimization_problem.add_variable("eggs", lb=0, ub=1)

        optimization_problem.add_linear_constraint(["foo", "spam"], [-1, 1])
        optimization_problem.add_linear_constraint(["foo", "eggs"], [-1, 1])
        optimization_problem.add_linear_constraint(["eggs", "spam"], [-1, 1])

        def copy_var(var):
            return var

        optimization_problem.add_variable_dependency("spam", "bar", copy_var)

        self.optimization_problem = optimization_problem

    def test_dependencies(self):
        independent_variables_expected = ["foo", "bar", "eggs"]
        independent_variables = self.optimization_problem.independent_variable_names
        self.assertEqual(independent_variables_expected, independent_variables)

        def transform():
            pass

        # Variable does not exist
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable_dependency(
                "inexistent", ["bar", "spam"], transform
            )

        # Dependent Variable does not exist
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable_dependency(
                "foo", ["inexistent", "spam"], transform
            )

        # Variable is already dependent
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable_dependency(
                "spam", ["bar", "spam"], transform
            )

        # Transform is not callable
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable_dependency(
                "spam", ["bar", "spam"], transform=None
            )

        independent_variables = self.optimization_problem.independent_variable_names
        self.assertEqual(independent_variables_expected, independent_variables)

        dependent_variables_expected = ["spam"]
        dependent_variables = self.optimization_problem.dependent_variable_names
        self.assertEqual(dependent_variables_expected, dependent_variables)

        variables_expected = ["foo", "bar", "spam", "eggs"]
        variables = self.optimization_problem.variable_names
        self.assertEqual(variables_expected, variables)

    def test_initial_values_without_dependencies(self):
        x0_chebyshev_expected = [0.70710678, 0.29289322, 0.29289322]
        x0_chebyshev = self.optimization_problem.get_chebyshev_center(
            include_dependent_variables=False
        )
        np.testing.assert_almost_equal(x0_chebyshev, x0_chebyshev_expected)

        variables_expected = [0.70710678, 0.29289322, 0.29289322, 0.29289322]
        variables = self.optimization_problem.get_dependent_values(x0_chebyshev)
        np.testing.assert_almost_equal(variables, variables_expected)

        x0_seed_1_expected = [[0.8324936, 0.279713, 0.6649623]]
        x0_seed_1 = self.optimization_problem.create_initial_values(
            1, seed=1, include_dependent_variables=False
        )
        np.testing.assert_almost_equal(x0_seed_1, x0_seed_1_expected)
        self.assertTrue(
            self.optimization_problem.check_linear_constraints(
                x0_seed_1[0], get_dependent_values=True
            )
        )

        x0_seed_1_random = self.optimization_problem.create_initial_values(
            1, include_dependent_variables=False
        )

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_1_random, x0_seed_1_expected)

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_1_random, x0_chebyshev_expected)

        x0_seed_10_expected = [
            [0.8324936, 0.27971297, 0.66496231],
            [0.73128644, 0.17275154, 0.5115499],
            [0.95771264, 0.19106333, 0.79216119],
            [0.81089326, 0.29062957, 0.6673637],
            [0.76359474, 0.20575487, 0.59221373],
            [0.9596961, 0.6718687, 0.90182574],
            [0.89945427, 0.01653708, 0.49987409],
            [0.92682306, 0.07106034, 0.8258856],
            [0.77868893, 0.4824452, 0.77038195],
            [0.59375802, 0.03129941, 0.22275326],
        ]
        x0_seed_10 = self.optimization_problem.create_initial_values(
            10, seed=1, include_dependent_variables=False
        )
        np.testing.assert_almost_equal(x0_seed_10, x0_seed_10_expected)

        x0_seed_10_random = self.optimization_problem.create_initial_values(
            10, include_dependent_variables=False
        )

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_10_random, x0_seed_10_expected)

    def test_initial_values(self):
        x0_chebyshev_expected = [0.70710678, 0.29289322, 0.29289322, 0.29289322]
        x0_chebyshev = self.optimization_problem.get_chebyshev_center(
            include_dependent_variables=True
        )
        np.testing.assert_almost_equal(x0_chebyshev, x0_chebyshev_expected)

        independent_variables_expected = [0.70710678, 0.29289322, 0.29289322]
        independent_variables = self.optimization_problem.get_independent_values(
            x0_chebyshev
        )
        np.testing.assert_almost_equal(
            independent_variables, independent_variables_expected
        )

        x0_seed_1_expected = [[0.8324936, 0.279713, 0.279713, 0.6649623]]
        x0_seed_1 = self.optimization_problem.create_initial_values(
            1, seed=1, include_dependent_variables=True
        )
        np.testing.assert_almost_equal(x0_seed_1, x0_seed_1_expected)
        self.assertTrue(
            self.optimization_problem.check_linear_constraints(x0_seed_1[0])
        )

        x0_seed_1_random = self.optimization_problem.create_initial_values(
            1, include_dependent_variables=True
        )[0]

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_1_random, x0_seed_1_expected)

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_1_random, x0_chebyshev_expected)

        x0_seed_10_expected = [
            [0.8324936, 0.27971297, 0.27971297, 0.66496231],
            [0.73128644, 0.17275154, 0.17275154, 0.5115499],
            [0.95771264, 0.19106333, 0.19106333, 0.79216119],
            [0.81089326, 0.29062957, 0.29062957, 0.6673637],
            [0.76359474, 0.20575487, 0.20575487, 0.59221373],
            [0.9596961, 0.6718687, 0.6718687, 0.90182574],
            [0.89945427, 0.01653708, 0.01653708, 0.49987409],
            [0.92682306, 0.07106034, 0.07106034, 0.8258856],
            [0.77868893, 0.4824452, 0.4824452, 0.77038195],
            [0.59375802, 0.03129941, 0.03129941, 0.22275326],
        ]
        x0_seed_10 = self.optimization_problem.create_initial_values(
            10, seed=1, include_dependent_variables=True
        )
        np.testing.assert_almost_equal(x0_seed_10, x0_seed_10_expected)

        x0_seed_10_random = self.optimization_problem.create_initial_values(
            10, include_dependent_variables=True
        )

        with self.assertRaises(AssertionError):
            np.testing.assert_almost_equal(x0_seed_10_random, x0_seed_10_expected)


class Test_OptimizationProblemJacobian(unittest.TestCase):
    def setUp(self):
        def single_obj_single_var(x):
            return x[0] ** 2

        name = "single_obj_single_var"
        optimization_problem = OptimizationProblem(name, use_diskcache=False)
        optimization_problem.add_variable("x")
        optimization_problem.add_objective(single_obj_single_var)
        self.single_obj_single_var = optimization_problem

        def single_obj_two_vars(x):
            return x[1] ** 2 - x[0] ** 2

        name = "single_obj_two_vars"
        optimization_problem = OptimizationProblem(name, use_diskcache=False)
        optimization_problem.add_variable("x_1")
        optimization_problem.add_variable("x_2")
        optimization_problem.add_objective(single_obj_two_vars)
        self.single_obj_two_vars = optimization_problem

        def two_obj_single_var(x):
            return [x[0] ** 2, x[0] ** 2]

        name = "two_obj_single_var"
        optimization_problem = OptimizationProblem(name, use_diskcache=False)
        optimization_problem.add_variable("x")
        optimization_problem.add_objective(two_obj_single_var, n_objectives=2)
        self.two_obj_single_var = optimization_problem

        def two_obj_two_var(x):
            return [x[0] ** 2, x[1] ** 2]

        name = "two_obj_two_var"
        optimization_problem = OptimizationProblem(name, use_diskcache=False)
        optimization_problem.add_variable("x_1")
        optimization_problem.add_variable("x_2")
        optimization_problem.add_objective(two_obj_two_var, n_objectives=2)
        self.two_obj_two_var = optimization_problem

    def test_jacobian(self):
        jac_expected = [[4.001]]
        jac = self.single_obj_single_var.objective_jacobian([2])
        np.testing.assert_almost_equal(jac, jac_expected)

        jac_expected = [[0.001]]
        jac = self.single_obj_single_var.objective_jacobian([0])
        np.testing.assert_almost_equal(jac, jac_expected)

        jac_expected = [[-4.001, 4.001]]
        jac = self.single_obj_two_vars.objective_jacobian([2, 2])
        np.testing.assert_almost_equal(jac, jac_expected)

        jac_expected = [[-0.001, 0.001]]
        jac = self.single_obj_two_vars.objective_jacobian([0, 0])
        np.testing.assert_almost_equal(jac, jac_expected)

        jac_expected = [[4.001], [4.001]]
        jac = self.two_obj_single_var.objective_jacobian([2])
        np.testing.assert_almost_equal(jac, jac_expected)

        jac_expected = [[0.001], [0.001]]
        jac = self.two_obj_single_var.objective_jacobian([0])
        np.testing.assert_almost_equal(jac, jac_expected)

        jac_expected = [[4.001, 0], [0, 4.001]]
        jac = self.two_obj_two_var.objective_jacobian([2, 2])
        np.testing.assert_almost_equal(jac, jac_expected)

        jac_expected = [[0.001, 0], [0, 0.001]]
        jac = self.two_obj_two_var.objective_jacobian([0, 0])
        np.testing.assert_almost_equal(jac, jac_expected)


class Test_OptimizationProblemConstraintTransforms(unittest.TestCase):
    """
    for linear transformation of constraints in an `OptimizationProblem`,
    tests if `A_transformed` and `b_transformed` properties are correctly
    computed.
    """

    def setup(self):
        settings.working_directory = "./test_problem"

    def tearDown(self):
        shutil.rmtree("./test_problem", ignore_errors=True)
        settings.working_directory = None

    @staticmethod
    def check_inequality_constraints(X, problem, transformed_space=False):
        """
        evaluates constraints in transformed space. This behavior is using
        the `A_transformed` and `b_transformed` properties of
        `OptimizationProblem`
        """

        if transformed_space:
            A = problem.A_transformed
            b = problem.b_transformed
        else:
            A = problem.A
            b = problem.b

        evaluate_constraints = lambda x: A.dot(x) - b

        lhs = np.array(list(map(evaluate_constraints, X)))
        rhs = 0
        CV = np.all(lhs <= rhs, axis=1)

        return CV

    @staticmethod
    def check_equality_constraints(X, problem, transformed_space=False):
        """
        evaluates constraints in transformed space. This behavior is using
        the `A_transformed` and `b_transformed` properties of
        `OptimizationProblem`
        """

        if transformed_space:
            Aeq = problem.Aeq_transformed
            beq = problem.beq_transformed
        else:
            Aeq = problem.Aeq
            beq = problem.beq

        evaluate_constraints = lambda x: Aeq.dot(x) - beq

        lhs = np.array(list(map(evaluate_constraints, X)))
        rhs = problem.eps_lineq
        CV = np.all(np.abs(lhs) <= rhs, axis=1)

        return CV

    @staticmethod
    def check_constraint_transform(problem, check_constraint_func):
        nvars = problem.n_independent_variables

        rng = np.random.default_rng(seed=72729)
        X = rng.uniform(0, 1, size=(100000, nvars))

        CV = check_constraint_func(X=X, problem=problem, transformed_space=True)

        # extract valid X and untransform then check constraints in
        # untransformed space
        X_valid = X[CV, :]
        X_valid = problem.untransform(X_valid)

        CV_test_valid = check_constraint_func(
            X=X_valid, problem=problem, transformed_space=False
        )

        # extract invalid X and untransform, then check constraints in
        # untransformed space
        X_invalid = X[~CV, :]
        X_invalid = problem.untransform(X_invalid)

        CV_test_invalid = check_constraint_func(
            X=X_invalid, problem=problem, transformed_space=False
        )

        # tests if all valid X remain valid in untransformed space
        assert np.all(CV_test_valid)

        # tests if all invalid X remain invalid in untransformed space
        assert np.all(~CV_test_invalid)

    def test_linear_inequality_constrained_transform(self):
        problem = LinearConstraintsSooTestProblem2(
            transform="linear",
            use_diskcache=False,
        )

        self.check_constraint_transform(problem, self.check_inequality_constraints)

    def test_linear_equality_constrained_transform(self):
        problem = LinearEqualityConstraintsSooTestProblem(
            transform="linear",
            use_diskcache=False,
        )

        self.check_constraint_transform(problem, self.check_equality_constraints)


class Test_OptimizationProblemEvaluator(unittest.TestCase):
    """
    Szenarien für Variablen:
        - Bound Constraints
        - Linear Constraints
        - Dependencies
        - Dependencies with Constraints

    Szenarien für EvaluationObjects:
        - Verschiedene Variablen für verschiedene EvalObjects.
        - Verschiedene Objectives/Constraints für verschiedene EvalObjects.
        - Verschiedene Evaluatoren für verschiedene Objectives/Constraints.
        - Caching nur für ausgewählte Evaluatoren.

    """

    def setUp(self):
        eval_obj = EvaluationObject()

        # Simple case
        optimization_problem = OptimizationProblem(
            "with_evaluator", use_diskcache=False
        )

        optimization_problem.add_evaluation_object(eval_obj)

        optimization_problem.add_variable("scalar_param")
        optimization_problem.add_variable("sized_list_param", lb=0, ub=10, indices=0)

        self.optimization_problem = optimization_problem

        def copy_var(var):
            return var

        optimization_problem.add_variable_dependency(
            "sized_list_param", "scalar_param", copy_var
        )

    def test_variable_names(self):
        names_expected = ["scalar_param", "sized_list_param"]
        names = self.optimization_problem.variable_names
        self.assertEqual(names_expected, names)

        # Variable does not exist in Evaluator
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable("bar", lb=0, ub=1)

        # Check that adding dummy variables still works
        self.optimization_problem.add_variable(
            "bar", evaluation_objects=None, lb=0, ub=1
        )
        names_expected = ["scalar_param", "sized_list_param", "bar"]
        names = self.optimization_problem.variable_names
        self.assertEqual(names_expected, names)

    def test_set_variables(self):
        pass

    def test_indices(self):
        pass

    def test_duplicate_variables(self):
        self.optimization_problem.check_duplicate_variables()

        # Duplicate name
        self.optimization_problem.add_variable("foo", evaluation_objects=None)

        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable("foo", evaluation_objects=None)

        # Duplicate scalar parameter variable
        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable(
                name="another_scalar_parameter",
                parameter_path="scalar_param",
            )

        # Duplicate index parameter variable
        self.optimization_problem.add_variable(
            name="another_sized_list_param_var",
            parameter_path="sized_list_param",
            lb=0,
            ub=10,
            indices=1,
        )

        with self.assertRaises(CADETProcessError):
            self.optimization_problem.add_variable(
                name="yet_another_sized_list_param_var",
                parameter_path="sized_list_param",
                lb=0,
                ub=10,
                indices=1,
            )

    def test_cache(self):
        pass


class Test_MultiEvaluationObjects(unittest.TestCase):
    def setUp(self):
        eval_obj_1 = EvaluationObject(name="foo")
        eval_obj_2 = EvaluationObject(name="bar")

        # Simple case
        optimization_problem = OptimizationProblem(
            "with_evaluator", use_diskcache=False
        )

        optimization_problem.add_evaluation_object(eval_obj_1)
        optimization_problem.add_evaluation_object(eval_obj_2)

        optimization_problem.add_variable(
            name="scalar_eval_obj_1",
            parameter_path="scalar_param",
            lb=0,
            ub=1,
            evaluation_objects=[eval_obj_1],
        )
        optimization_problem.add_variable(
            name="scalar_eval_obj_2",
            parameter_path="scalar_param",
            lb=0,
            ub=1,
            evaluation_objects=[eval_obj_2],
        )

        optimization_problem.add_variable(
            name="scalar_both_eval_obj",
            parameter_path="scalar_param_2",
            lb=0,
            ub=1,
        )

        def single_obj_1(eval_obj):
            return 0

        optimization_problem.add_objective(single_obj_1)

        def single_obj_2(eval_obj):
            return 1

        optimization_problem.add_objective(single_obj_2, evaluation_objects=eval_obj_1)

        def multi_obj(eval_obj):
            return [2, 3]

        optimization_problem.add_objective(multi_obj, n_objectives=2)

        self.optimization_problem = optimization_problem

    def test_evaluation(self):
        f_expected = [0, 0, 1, 2, 3, 2, 3]
        f = self.optimization_problem.evaluate_objectives([1, 1, 1])
        np.testing.assert_allclose(f, f_expected)

        f_expected = [[0, 0, 1, 2, 3, 2, 3], [0, 0, 1, 2, 3, 2, 3]]
        f = self.optimization_problem.evaluate_objectives([[1, 1, 1], [1, 1, 1]])

        np.testing.assert_allclose(f, f_expected)

    def test_names(self):
        names_expected = ["single_obj_1", "single_obj_2", "multi_obj"]
        names = self.optimization_problem.objective_names
        self.assertEqual(names, names_expected)

    def test_labels(self):
        labels_expected = [
            "foo_single_obj_1",
            "bar_single_obj_1",
            "single_obj_2",
            "foo_multi_obj_0",
            "bar_multi_obj_0",
            "foo_multi_obj_1",
            "bar_multi_obj_1",
        ]
        labels = self.optimization_problem.objective_labels
        self.assertEqual(labels, labels_expected)


if __name__ == "__main__":
    unittest.main()



"""
===========================================
Equilibria (:mod:`CADETProcess.equilibria`)
===========================================

.. currentmodule:: CADETProcess.equilibria

A collection of tools to calculate reaction and binding equilibria.

Buffer Capacity
===============

Calculate buffer capacity

.. autosummary::
   :toctree: generated/

   buffer_capacity

Reaction Equilibria
===================

Calculate consistent initial conditions of reaction systems.

.. autosummary::
   :toctree: generated/

   reaction_equilibria

Initial Conditions
==================

Calculate consistent initial conditions of reaction/binding system.

.. autosummary::
   :toctree: generated/

   initial_conditions

"""

from .ptc import *
from .reaction_equilibria import *
from .initial_conditions import *
from .buffer_capacity import *



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# (clr_process)=
# # Closed Loop Recycling Process
# In closed-loop recycling (CLR), the stock mixture is pumped over the column several times until the desired purity is achieved.
# The general structure of a CLR is shown below.
#
# ```{figure} ./figures/clr_flow_sheet.svg
# :name: clr_flow_sheet
#
# Flow sheet for closed-loop recycling process.
# ```
#
# To realize the recycling, the {attr}`~CADETProcess.processModel.FlowSheet.output_state` of the column needs to be modified, leading to the following event structure:
#
# ```{figure} ./figures/clr_events.svg
# :name: clr_events
#
# Events for closed-loop recycling process.
# ```
#
# For this example, consider a two-component system with a Langmuir isotherm.
#
# ## Component System

# %%
from CADETProcess.processModel import ComponentSystem

component_system = ComponentSystem(['A', 'B'])

# %% [markdown]
# ## Binding Model

# %%
from CADETProcess.processModel import Langmuir

binding_model = Langmuir(component_system, name='langmuir')
binding_model.adsorption_rate = [0.04, 0.05]
binding_model.desorption_rate = [1, 1]
binding_model.capacity = [100, 100]

# %% [markdown]
# ## Unit Operations

# %%
from CADETProcess.processModel import (
    Inlet, Cstr, LumpedRateModelWithoutPores, Outlet
)
feed = Inlet(component_system, name='feed')
feed.c = [10, 10]

eluent = Inlet(component_system, name='eluent')
eluent.c = [0, 0]

pump = Cstr(component_system, name='pump')
pump.V = 1e-9

column = LumpedRateModelWithoutPores(component_system, name='column')
column.binding_model = binding_model
column.length = 0.6
column.diameter = 0.024
column.axial_dispersion = 4.7e-7
column.total_porosity = 0.7

outlet = Outlet(component_system, name='outlet')

# %% [markdown]
# ## Flow Sheet

# %%
from CADETProcess.processModel import FlowSheet

flow_sheet = FlowSheet(component_system)

flow_sheet.add_unit(feed, feed_inlet=True)
flow_sheet.add_unit(eluent, eluent_inlet=True)
flow_sheet.add_unit(pump)
flow_sheet.add_unit(column)
flow_sheet.add_unit(outlet, product_outlet=True)

flow_sheet.add_connection(feed, column)
flow_sheet.add_connection(eluent, column)
flow_sheet.add_connection(column, outlet)
flow_sheet.add_connection(column, pump)
flow_sheet.add_connection(pump, column)

# %% [markdown]
# ## Process

# %%
from CADETProcess.processModel import Process
process = Process(flow_sheet, 'clr')

# %% [markdown]
# ### Create Events and Durations

# %%
Q = 60 / (60 * 1e6)

process.add_event('feed_on', 'flow_sheet.feed.flow_rate', Q)
process.add_event('feed_off', 'flow_sheet.feed.flow_rate', 0.0)

process.add_event('eluent_off', 'flow_sheet.eluent.flow_rate', 0.0)
process.add_event('eluent_on', 'flow_sheet.eluent.flow_rate', Q)

process.add_event('recycle_on_state', 'flow_sheet.output_states.column', {'pump': 1})
process.add_event('recycle_on_pump', 'flow_sheet.pump.flow_rate', Q)
process.add_event('recycle_off_state', 'flow_sheet.output_states.column', {'outlet': 1})
process.add_event('recycle_off_pump', 'flow_sheet.pump.flow_rate', 0)

# %% [markdown]
# ### Event dependencies
# To reduce the number of event times that need to be specified, event dependencies are specified which enforce that always either feed or eluent are being pumped through the column.

# %%
process.add_event_dependency('eluent_off', ['feed_on'])

process.add_event_dependency('recycle_on_state', ['feed_off'])
process.add_event_dependency('recycle_on_pump', ['feed_off'])

process.add_event_dependency('recycle_off_pump', ['recycle_off_state'])
process.add_event_dependency('eluent_on', ['recycle_off_state'])

# %% [markdown]
# ### Event Times
# Now, the cycle time is set to $10~min$ and the `feed_duration` to $1~min$.

# %%
process.cycle_time = 2000
process.feed_off.time = 40
process.recycle_off_state.time = 1280

# %% [markdown]
# ## Simulate Process
#
# Here, the first plot shows the concentration profile at the column outlet.
# It is important to note that since part of this profile is recycled, the concentration profile at the system outlet must be considered (second plot) to evaluate the process performance.

# %%
if __name__ == '__main__':
    from CADETProcess.simulator import Cadet
    process_simulator = Cadet()

    simulation_results = process_simulator.simulate(process)
    simulation_results.solution.column.outlet.plot()
    simulation_results.solution.outlet.inlet.plot()

# %% [markdown]
# ## Optimize Fractionation Times

# %%
if __name__ == '__main__':
    from CADETProcess.fractionation import FractionationOptimizer
    fractionation_optimizer = FractionationOptimizer()

    fractionator = fractionation_optimizer.optimize_fractionation(
        simulation_results, purity_required=[0.95, 0.95]
    )

    print(fractionator.performance)
    _ = fractionator.plot_fraction_signal()

# %% [markdown]
# ## Peak Shaving
# The disadvantage of the CLR process is an increased dispersion due to multiple passes through the pump and additional piping.
#
# To improve the overall process performance, the CLR process is often combined with peak shaving.
# In this process, the initial and final regions of the chromatogram with sufficient purity are "shaved off" during each cycle.
# Peak shaving can reduce the number of recycling cycles required, since a decreasing amount of components must be pumped across the column.
#
# ```{figure} ./figures/clr_peak_shaving_events.svg
# :name: clr_peak_shaving_events
#
# Events for closed-loop recycling process with peak shaving.
# ```



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% tags=["remove-cell"]
from pathlib import Path
import sys

root_dir = Path('../../../../').resolve()
sys.path.append(root_dir.as_posix())

# %% [markdown]
# # The Yamamoto method
#
# This example demonstrates how to estimate SMA binding parameters based on multiple gradient elution chromatograms
# using the Yamamoto method.

# %%
import numpy as np

from CADETProcess.processModel import ComponentSystem
from CADETProcess.tools.yamamoto import GradientExperiment, fit_parameters

from binding_model_parameters import create_column_model, create_in_silico_experimental_data

if __name__ == "__main__":
    component_system = ComponentSystem(['Salt', 'Protein'])
    column = create_column_model(component_system, final_salt_concentration=600, initial_salt_concentration=50)

    column_volume = column.length * ((column.diameter / 2) ** 2) * np.pi

    create_in_silico_experimental_data()

    exp_5cv = np.loadtxt("experimental_data/5.csv", delimiter=",")
    exp_30cv = np.loadtxt("experimental_data/30.csv", delimiter=",")
    exp_120cv = np.loadtxt("experimental_data/120.csv", delimiter=",")

    experiment_1 = GradientExperiment(exp_5cv[:, 0], exp_5cv[:, 1], exp_5cv[:, 2], 5 * column_volume)
    experiment_2 = GradientExperiment(exp_30cv[:, 0], exp_30cv[:, 1], exp_30cv[:, 2], 30 * column_volume)
    experiment_3 = GradientExperiment(exp_120cv[:, 0], exp_120cv[:, 1], exp_120cv[:, 2], 120 * column_volume)

    experiments = [experiment_1, experiment_2, experiment_3]

    for experiment in experiments:
        experiment.plot()

    yamamoto_results = fit_parameters(experiments, column)

    print('yamamoto_results.characteristic_charge =', yamamoto_results.characteristic_charge)
    print('yamamoto_results.k_eq =', yamamoto_results.k_eq)

    yamamoto_results.plot()



import warnings

import numba
import numpy as np
import numpy.typing as npt
import scipy
from scipy.interpolate import PchipInterpolator

from CADETProcess import CADETProcessError


def linear_coeff(x1: float, y1: float, x2: float, y2: float) -> tuple[float, float]:
    """Return paramters that fit y = a*x+b."""
    a = (y2 - y1) / (x2 - x1)
    b = y2 - a * x2
    return a, b


def exponential_coeff(
    x1: float, y1: float, x2: float, y2: float
) -> tuple[float, float]:
    """Return paramaters that fit y = b*exp(m*x)."""
    b = (np.log(y2) - np.log(y1)) / (x2 - x1)
    a = y1 * np.exp(-b * x1)
    return a, b


def exponential(x: float, a: float, b: float) -> float:
    """Evaluate exponential function."""
    return a * np.exp(b * x)


def linear(x: float, a: float, b: float) -> float:
    """Evaluate linear function."""
    return a * x + b


def find_opt_poly(
    x: np.ndarray,
    y: np.ndarray,
    index: int,
) -> tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Find optimal overlap between x (offset) and y (pearson).

    Given a curve, find highest point.
    """
    if index == 0:
        indices = np.array([index, index + 1, index + 2])
    elif index == (len(x) - 1):
        indices = np.array([index - 2, index - 1, index])
    else:
        indices = np.array([index - 1, index, index + 1])

    x = x[indices]
    y = y[indices]

    if x[0] > x[-1]:  # need to invert order
        x = x[::-1]
        y = y[::-1]

    try:
        poly, res = np.polynomial.Polynomial.fit(x, y, 2, full=True)
    except np.linalg.LinAlgError:
        raise CADETProcessError(f"Polyfit failed for {x}, {y}")
    try:
        root = poly.deriv().roots()[0]
    except IndexError:
        # If all y values are equal, use center value.
        root = x[0]
    root = np.clip(root, x[0], x[1])
    return root, x, y


def pear_corr(cr: float) -> float:
    """Flip pearson correlation s.t. 0 is best and 1 is worst."""
    # handle the case where a nan is returned
    if np.isnan(cr):
        return 1.0
    if cr < 0.0:
        return 1.0
    else:
        return 1 - cr


@numba.njit(fastmath=True)
def pearsonr_mat(x: np.ndarray, Y: np.ndarray, times: np.ndarray) -> np.ndarray:
    """
    High performance implementation of the pearson correlation.

    This is to simultaneously evaluate the pearson correlation between a vector and a
    matrix. Scipy can only evaluate vector/vector.
    """
    r = np.zeros(Y.shape[0])
    xm = x - x.mean()

    r_x_den = np.linalg.norm(xm)

    for i in range(Y.shape[0]):
        ym = Y[i] - np.mean(Y[i])

        r_num = np.dot(xm, ym)
        r_y_den = np.linalg.norm(ym)

        denominator = r_x_den * r_y_den

        if denominator == 0.0:
            r[i] = -1.0
        else:
            min_fun = 0
            for j in range(x.shape[0]):
                min_fun += min(x[j], Y[i, j])

            r[i] = min(max(r_num / denominator, -1.0), 1.0) * min_fun
    return r


def pearson_offset(
    time: npt.ArrayLike,
    reference_spline: PchipInterpolator,
    simulation_spline: PchipInterpolator,
    offset: float,
) -> float:
    """Calculate single pearson correlation at offset."""
    simulation_data_offset = simulation_spline(time - offset)
    reference_data = reference_spline(time)
    try:
        pear = scipy.stats.pearsonr(reference_data, simulation_data_offset)[0]
    except ValueError:
        warnings.warn(
            f"Pearson correlation failed to do NaN or InF in array exp_array: "
            f"{list(reference_data)}, sim_array: {list(simulation_data_offset)}"
        )
        pear = 0
    score_local = pear_corr(pear)

    return score_local


def eval_offsets(
    time: float,
    reference_spline: PchipInterpolator,
    simulation_spline: PchipInterpolator,
    offsets: np.ndarray,
) -> np.ndarray:
    """Calculate pearson correlation for each offset."""
    rol_mat = np.zeros([len(offsets), len(time)])

    for idx, offset in enumerate(offsets):
        rol_mat[idx, :] = simulation_spline(time - offset)

    reference_data = reference_spline(time)
    scores = pearsonr_mat(reference_data, rol_mat, time)
    return scores


def pearson(
    time: np.ndarray,
    reference_spline: PchipInterpolator,
    simulation_spline: PchipInterpolator,
    size: int = 20,
    nest: int = 50,
    bounds: int = 2,
    tol: float = 1e-13,
) -> tuple[float, float]:
    """
    Find highest correlation between reference and simulation.

    The two signals are shifted in time to find the time offset which
    corresponds to highest correlation. To ensure deterministic results, a
    a bisection scheme is used to refine the offset instead of an optimizer.

    Parameters
    ----------
    time: np.ndarray
        Time points.
    reference_spline : PchipInterpolator
        Reference data points.
    simulation_spline : PchipInterpolator
        Simulation Data points.
    size : int, optional
        Number of points to be generated between upper and lower bounds.
        The default is 20.
    nest : int, optional
        Maximum number of recursions. The default is 50.
    bounds : int, optional
        Number of indices allowed for shifting signal outside bounds.
        The default is 2.
    tol : float, optional
        Tolerance for difference between lb and ub. The default is 1e-13.

    Returns
    -------
    score_local : float
        pearson correlation (inverted).
    dt : float
        time offset.
    """
    for i in range(nest + 1):
        if i == 0:
            lb = -time[-1]
            ub = time[-1]
            local_size = min(100 + 1, int((ub - lb) * 2 + 1))
        else:
            idx_max = np.argmax(pearson)

            try:
                lb = offsets[idx_max - bounds]  # noqa: F821
            except IndexError:
                lb = offsets[0]  # noqa: F821

            try:
                ub = offsets[idx_max + bounds]  # noqa: F821
            except IndexError:
                ub = offsets[-1]  # noqa: F821
            local_size = size

        if ub - lb < tol:
            break

        offsets = np.linspace(lb, ub, local_size)

        pearson = eval_offsets(time, reference_spline, simulation_spline, offsets)

        idx_max = np.argmax(pearson)

        expand_lb = max(bounds - idx_max, 0)
        expand_ub = max(bounds - (len(pearson) - 1 - idx_max), 0)

        if expand_lb or expand_ub:
            expand_lb = expand_lb * 2
            expand_ub = expand_ub * 2
            dt = offsets[1] - offsets[0]
            if expand_lb:
                local_offsets = np.linspace(offsets[0] - expand_lb * dt, offsets[0] - dt, expand_lb)
                local_pearson = eval_offsets(
                    time, reference_spline, simulation_spline, local_offsets
                )

                offsets = np.concatenate([local_offsets, offsets])
                pearson = np.concatenate([local_pearson, pearson])

            if expand_ub:
                local_offsets = np.linspace(
                    offsets[-1] + dt, offsets[-1] + expand_ub * dt, expand_ub
                )

                local_pearson = eval_offsets(
                    time, reference_spline, simulation_spline, local_offsets
                )

                offsets = np.concatenate([offsets, local_offsets])
                pearson = np.concatenate([pearson, local_pearson])

    idx = np.argmax(pearson)

    dt, time_found, goal_found = find_opt_poly(offsets, pearson, idx)

    # calculate pearson correlation at the new time
    score_local = pearson_offset(time, reference_spline, simulation_spline, dt)

    return score_local, dt



import unittest

import numpy as np
from CADETProcess.dynamicEvents import MultiTimeLine, Section, TimeLine
from CADETProcess.dynamicEvents.section import generate_indices


class TestGenerateIndices(unittest.TestCase):
    def test_generate_indices(self):
        shape = (3, 3)

        indices = [[0, 1], [1, 2]]
        indices_tuple_expected = [(0, 1), (1, 2)]
        indices_tuple = generate_indices(shape, indices)
        np.testing.assert_equal(indices_tuple, indices_tuple_expected)

        indices = np.s_[:]
        indices_tuple_expected = [(slice(None, None, None),)]
        indices_tuple = generate_indices(shape, indices)
        np.testing.assert_equal(indices_tuple, indices_tuple_expected)

        indices = np.s_[0, :]
        indices_tuple_expected = [(0, slice(None, None, None))]
        indices_tuple = generate_indices(shape, indices)
        np.testing.assert_equal(indices_tuple, indices_tuple_expected)

        indices = [np.s_[0, :], [1, 1]]
        indices_tuple_expected = [(0, slice(None, None, None)), (1, 1)]
        indices_tuple = generate_indices(shape, indices)
        np.testing.assert_equal(indices_tuple, indices_tuple_expected)

        with self.assertRaises(IndexError):
            _ = generate_indices(shape, indices=[3, 3])

        with self.assertRaises(ValueError):
            _ = generate_indices((), indices=[[0, 1], [1, 2]])


class TestSection(unittest.TestCase):
    def setUp(self):
        self.constant_section_single = Section(0, 1, 1)
        self.constant_section_multi = Section(1, 2, [1, 2])
        self.poly_section_single = Section(0, 1, [0, 1, 0, 0], is_polynomial=True)
        self.poly_section_multi = Section(0, 1, [[0, 1], [1, -1]], is_polynomial=True)

    def test_coeffs(self):
        np.testing.assert_equal(self.constant_section_single.coeffs, [[1]])
        np.testing.assert_equal(self.constant_section_multi.coeffs, [[1], [2]])
        np.testing.assert_equal(self.poly_section_single.coeffs, [[0, 1, 0, 0]])
        np.testing.assert_equal(self.poly_section_multi.coeffs, [[0, 1], [1, -1]])

    def test_section_value(self):
        const_single = self.constant_section_single
        self.assertEqual(const_single.value(0), 1)

        # Exceed section times
        with self.assertRaises(ValueError):
            val = const_single.value(2)

        const_multi = self.constant_section_multi
        np.testing.assert_equal(const_multi.value(1), [1, 2])

        # Exceed section times
        with self.assertRaises(ValueError):
            val = const_multi.value(0)

        poly_single = self.poly_section_single
        np.testing.assert_equal(poly_single.value(0), 0)
        np.testing.assert_equal(poly_single.value(0.5), 0.5)
        np.testing.assert_equal(poly_single.value(1), 1)

        poly_multi = self.poly_section_multi
        np.testing.assert_equal(poly_multi.value(0), [0, 1])
        np.testing.assert_equal(poly_multi.value(0.5), [0.5, 0.5])
        np.testing.assert_equal(poly_multi.value(1), [1, 0])

    def test_section_integral(self):
        const_single = self.constant_section_single
        self.assertEqual(const_single.integral(0, 0), 0)
        self.assertEqual(const_single.integral(0, 1), 1)

        # Exceed section times
        with self.assertRaises(ValueError):
            val = const_single.value(2)

        const_multi = self.constant_section_multi
        np.testing.assert_equal(const_multi.integral(1, 2), [1, 2])

        # Exceed section times
        with self.assertRaises(ValueError):
            val = const_multi.value(0)

        poly_single = self.poly_section_single
        np.testing.assert_equal(poly_single.integral(0, 0.5), 0.125)
        np.testing.assert_equal(poly_single.integral(0, 1), 0.5)

        poly_multi = self.poly_section_multi
        np.testing.assert_equal(poly_multi.integral(0, 0.5), [0.125, 0.375])
        np.testing.assert_equal(poly_multi.integral(0.5, 1), [0.375, 0.125])
        np.testing.assert_equal(poly_multi.integral(0, 1), [0.5, 0.5])


class TestTimeLine(unittest.TestCase):
    def create_timeline_constant_single(self):
        """Piecewise constant sections with single entry."""
        section_0 = Section(0, 1, 1.5)
        section_1 = Section(1, 2, 0)
        section_2 = Section(2, 3, 0)
        section_3 = Section(3, 4, 1)
        section_4 = Section(4, 5, 2)
        section_5 = Section(5, 6, 2)

        tl = TimeLine()
        tl.add_section(section_0)
        tl.add_section(section_1)
        tl.add_section(section_2)
        tl.add_section(section_3)
        tl.add_section(section_4)
        tl.add_section(section_5)

        return tl

    def create_timeline_constant_multi(self):
        """Piecewise constant sections with multiple entries (differing in value)."""
        section_0 = Section(0, 1, (1.5, 0))
        section_1 = Section(1, 2, (0, 0))
        section_2 = Section(2, 3, (0, 1))
        section_3 = Section(3, 4, (1, 0))
        section_4 = Section(4, 5, (2, 0))
        section_5 = Section(5, 6, (2, -2))

        tl = TimeLine()
        tl.add_section(section_0)
        tl.add_section(section_1)
        tl.add_section(section_2)
        tl.add_section(section_3)
        tl.add_section(section_4)
        tl.add_section(section_5)

        return tl

    def create_timeline_poly_single(self):
        """Polynomial sections with single entry."""
        section_0 = Section(0, 1, (1.5, 0), is_polynomial=True)
        section_1 = Section(1, 2, (0, 0), is_polynomial=True)
        section_2 = Section(2, 3, (0, 1), is_polynomial=True)
        section_3 = Section(3, 4, (1, 0), is_polynomial=True)
        section_4 = Section(4, 5, (2, 0), is_polynomial=True)
        section_5 = Section(5, 6, (2, -2), is_polynomial=True)

        tl = TimeLine()
        tl.add_section(section_0)
        tl.add_section(section_1)
        tl.add_section(section_2)
        tl.add_section(section_3)
        tl.add_section(section_4)
        tl.add_section(section_5)

        return tl

    def create_timeline_poly_multi(self):
        """Polynomial sections with multiple entries (differing in value)."""
        section_0 = Section(0, 1, [[1.5, 0, 0], [0, 1, 0]], is_polynomial=True)
        section_1 = Section(1, 2, [[0, 0, 0], [1, 1, 0]], is_polynomial=True)
        section_2 = Section(2, 3, [[0, 1, 0], [1, 1, 0]], is_polynomial=True)
        section_3 = Section(3, 4, [[1, 0, 0], [0, 0, 0]], is_polynomial=True)
        section_4 = Section(4, 5, [[2, 0, 0], [0, 0, 0]], is_polynomial=True)
        section_5 = Section(5, 6, [[2, -2, 0], [0, 0, 1]], is_polynomial=True)

        tl = TimeLine()
        tl.add_section(section_0)
        tl.add_section(section_1)
        tl.add_section(section_2)
        tl.add_section(section_3)
        tl.add_section(section_4)
        tl.add_section(section_5)

        return tl

    def test_timeline_value(self):
        """Test values at specific times."""
        tl = self.create_timeline_constant_single()

        np.testing.assert_equal(tl.value(0), 1.5)
        np.testing.assert_equal(tl.value(1), 0.0)
        np.testing.assert_equal(tl.value(2), 0.0)
        np.testing.assert_equal(tl.value(3), 1.0)
        np.testing.assert_equal(tl.value(4), 2.0)
        np.testing.assert_equal(tl.value(5), 2.0)
        np.testing.assert_equal(tl.value(5.5), 2.0)
        np.testing.assert_equal(tl.value(6), 2.0)

        tl = self.create_timeline_constant_multi()

        np.testing.assert_equal(tl.value(0), [1.5, 0])
        np.testing.assert_equal(tl.value(1), [0.0, 0])
        np.testing.assert_equal(tl.value(2), [0.0, 1])
        np.testing.assert_equal(tl.value(3), [1.0, 0])
        np.testing.assert_equal(tl.value(4), [2.0, 0])
        np.testing.assert_equal(tl.value(5), [2.0, -2])
        np.testing.assert_equal(tl.value(5.5), [2.0, -2])
        np.testing.assert_equal(tl.value(6), [2.0, -2])

        tl = self.create_timeline_poly_single()

        np.testing.assert_equal(tl.value(2), 0.0)
        np.testing.assert_equal(tl.value(2.5), 0.5)
        np.testing.assert_equal(tl.value(5), 2.0)
        np.testing.assert_equal(tl.value(5.5), 1.0)
        np.testing.assert_equal(tl.value(6), 0.0)

        tl = self.create_timeline_poly_multi()
        np.testing.assert_equal(tl.value(0), [1.5, 0])
        np.testing.assert_equal(tl.value(2.5), [0.5, 1.5])
        np.testing.assert_equal(tl.value(5), [2.0, 0])
        np.testing.assert_equal(tl.value(5.5), [1.0, 0.25])
        np.testing.assert_equal(tl.value(6), [0.0, 1])

    def test_timeline_integral(self):
        """Test definite integral values for specific time intervals."""
        tl = self.create_timeline_poly_single()

        np.testing.assert_equal(tl.integral(0, 0), 0.0)
        np.testing.assert_equal(tl.integral(0, 0.5), 1.5 / 2)
        np.testing.assert_equal(tl.integral(0, 1), 1.5)
        np.testing.assert_equal(tl.integral(0, 2), 1.5)
        np.testing.assert_equal(tl.integral(2, 2.5), 0.5 / 2 / 2)
        np.testing.assert_equal(tl.integral(2, 3), 0.5)

        tl = self.create_timeline_poly_multi()

        np.testing.assert_equal(tl.integral(0, 0), [0.0, 0.0])
        np.testing.assert_equal(tl.integral(0, 0.5), [1.5 / 2, 0.125])
        np.testing.assert_equal(tl.integral(0, 1), [1.5, 0.5])
        np.testing.assert_equal(tl.integral(0, 2), [1.5, 2.0])
        np.testing.assert_equal(tl.integral(5, 6), [1, 1 / 3])

    def test_timeline_coeff(self):
        """Test coefficient values at given times."""
        tl = self.create_timeline_poly_single()
        np.testing.assert_equal(tl.coefficients(0.0), [1.5, 0])
        np.testing.assert_equal(tl.coefficients(5.5), [1, -2])

        tl = self.create_timeline_poly_multi()

        np.testing.assert_equal(tl.coefficients(0.0), [[1.5, 0, 0], [0, 1, 0]])

        np.testing.assert_equal(tl.coefficients(5.5), [[1, -2, 0], [0.25, 1, 1]])

    def test_section_times(self):
        """Test section times."""
        tl = self.create_timeline_constant_single()

        self.assertEqual(tl.section_times, [0, 1, 2, 3, 4, 5, 6])

    def test_tl_from_profile(self):
        """Test creation of time line from time series profile."""
        time = np.linspace(0, 100, 1001)
        y = np.sin(time / 10)

        tl = TimeLine.from_profile(time, y)
        np.testing.assert_almost_equal(tl.value(time)[:, 0], y, decimal=3)


class TestMultiTimeLine(unittest.TestCase):
    def create_timeline_constant_multi(self):
        """Piecewise constant sections with multiple entries managed by MultiTimeline."""
        section_0_0 = Section(0, 1, 1.5)
        section_0_1 = Section(1, 3, 0)
        section_0_2 = Section(3, 4, 1)
        section_0_3 = Section(4, 6, 2)

        section_1_0 = Section(0, 2, 0)
        section_1_1 = Section(2, 3, 1)
        section_1_2 = Section(3, 5, 0)
        section_1_3 = Section(5, 6, -2)

        tl = MultiTimeLine([0, 0])
        tl.add_section(section_0_0, 0)
        tl.add_section(section_0_1, 0)
        tl.add_section(section_0_2, 0)
        tl.add_section(section_0_3, 0)

        tl.add_section(section_1_0, 1)
        tl.add_section(section_1_1, 1)
        tl.add_section(section_1_2, 1)
        tl.add_section(section_1_3, 1)

        return tl

    def create_timeline_poly_multi(self):
        """Polynomial sections with multiple entries managed by MultiTimeline."""
        # Entry 0
        # Const. Coeff.
        section_0_0_0 = Section(0, 1, 1.5)
        section_0_0_1 = Section(1, 3, 0)
        section_0_0_2 = Section(3, 4, 1)
        section_0_0_3 = Section(4, 6, 2)

        # Lin. Coeff.
        section_0_1_0 = Section(0, 2, 0)
        section_0_1_1 = Section(2, 3, 1)
        section_0_1_2 = Section(3, 5, 0)
        section_0_1_3 = Section(5, 6, -2)

        # Entry 1
        # Const. Coeff.
        section_1_0_0 = Section(0, 1, 0)
        section_1_0_1 = Section(1, 3, 1)
        section_1_0_2 = Section(3, 6, 0)

        # Lin. Coeff.
        section_1_1_0 = Section(0, 3, 1)
        section_1_1_1 = Section(3, 6, 0)

        # Cubic Coeff
        section_1_2_0 = Section(0, 5, 0)
        section_1_2_1 = Section(5, 6, 1)

        tl = MultiTimeLine(base_state=[[0, 0, 0, 0], [0, 0, 0, 0]], is_polynomial=True)
        tl.add_section(section_0_0_0, entry_index=(0, 0))
        tl.add_section(section_0_0_1, entry_index=(0, 0))
        tl.add_section(section_0_0_2, entry_index=(0, 0))
        tl.add_section(section_0_0_3, entry_index=(0, 0))

        tl.add_section(section_0_1_0, entry_index=(0, 1))
        tl.add_section(section_0_1_1, entry_index=(0, 1))
        tl.add_section(section_0_1_2, entry_index=(0, 1))
        tl.add_section(section_0_1_3, entry_index=(0, 1))

        tl.add_section(section_1_0_0, entry_index=(1, 0))
        tl.add_section(section_1_0_1, entry_index=(1, 0))
        tl.add_section(section_1_0_2, entry_index=(1, 0))

        tl.add_section(section_1_1_0, entry_index=(1, 1))
        tl.add_section(section_1_1_1, entry_index=(1, 1))

        tl.add_section(section_1_2_0, entry_index=(1, 2))
        tl.add_section(section_1_2_1, entry_index=(1, 2))

        return tl

    # TODO: Test when one base state is not modified
    # TODO: Test exceeding index
    # TODO: Test missing sections

    def test_multi_timeline_value(self):
        multi_tl = self.create_timeline_constant_multi()
        tl = multi_tl.combined_time_line

        np.testing.assert_equal(tl.value(0), [1.5, 0])
        np.testing.assert_equal(tl.value(1), [0.0, 0])
        np.testing.assert_equal(tl.value(2), [0.0, 1])
        np.testing.assert_equal(tl.value(3), [1.0, 0])
        np.testing.assert_equal(tl.value(4), [2.0, 0])
        np.testing.assert_equal(tl.value(5), [2.0, -2])
        np.testing.assert_equal(tl.value(5.5), [2.0, -2])
        np.testing.assert_equal(tl.value(6), [2.0, -2])

        multi_tl = self.create_timeline_poly_multi()
        tl = multi_tl.combined_time_line

        np.testing.assert_equal(tl.value(0), [1.5, 0])
        np.testing.assert_equal(tl.value(2.5), [0.5, 2.5])
        np.testing.assert_equal(tl.value(5), [2.0, 0])
        np.testing.assert_equal(tl.value(5.5), [1.0, 0.25])
        np.testing.assert_equal(tl.value(6), [0.0, 1])

    def test_timeline_integral(self):
        """Test definite integral values for specific time intervals."""
        multi_tl = self.create_timeline_poly_multi()
        tl = multi_tl.combined_time_line

        np.testing.assert_equal(tl.integral(0, 0), [0.0, 0.0])
        np.testing.assert_equal(tl.integral(0, 0.5), [1.5 / 2, 0.125])
        np.testing.assert_equal(tl.integral(0, 1), [1.5, 0.5])
        np.testing.assert_equal(tl.integral(0, 2), [1.5, 2.0])
        np.testing.assert_equal(tl.integral(5, 6), [1, 1 / 3])

    def test_timeline_coeff(self):
        """Test coefficient values at given times."""
        multi_tl = self.create_timeline_poly_multi()
        tl = multi_tl.combined_time_line

        np.testing.assert_equal(tl.coefficients(0.0), [[1.5, 0, 0, 0], [0, 1, 0, 0]])
        np.testing.assert_equal(tl.coefficients(5.5), [[1, -2, 0, 0], [0.25, 1, 1, 0]])

    def test_section_times(self):
        """Test section times."""
        tl = self.create_timeline_constant_multi()

        self.assertEqual(tl.section_times, [0, 1, 2, 3, 4, 5, 6])

    def test_multi_timeline(self):
        base_state = [1, 2, 3]
        multi_tl_list = MultiTimeLine(base_state)
        # multi_tl_list.combined_time_line

        base_state = [[1, 2, 3], [4, 5, 6]]
        multi_tl_array = MultiTimeLine(base_state)
        # multi_tl_array.combined_time_line

        base_state = [1, 2, 3]
        multi_tl_list_poly = MultiTimeLine(base_state, is_polynomial=True)
        # multi_tl_list_poly.combined_time_line

        base_state = [[1, 2, 3], [4, 5, 6]]
        multi_tl_poly = MultiTimeLine(base_state, is_polynomial=True)
        # multi_tl_poly.combined_time_line

    def test_combined_timeline(self):
        pass


if __name__ == "__main__":
    unittest.main()



import shutil
import tempfile
from collections import defaultdict
from pathlib import Path
from typing import Any, Optional

from diskcache import Cache, FanoutCache

from CADETProcess.dataStructure import DillDisk


class ResultsCache:
    """
    Cache to store (intermediate) results.

    Optinally uses diskcache library to store large objects in sqlite database.

    Internal structure:
    [evaluation_object][step][x]

    For example:
    [EvaluationObject 1][Evaluator 1][x] -> IntermediateResults 1
    [EvaluationObject 1][Evaluator 2][x] -> IntermediateResults 2
    [EvaluationObject 1][Objective 1][x] -> f1.1
    [EvaluationObject 1][Objective 2][x] -> f1.2
    [EvaluationObject 1][Constraint 1][x] -> g1.1

    [EvaluationObject 2][Evaluator 1][x] -> IntermediateResults 1
    [EvaluationObject 2][Evaluator 2][x] -> IntermediateResults 2
    [EvaluationObject 2][Objective 1][x] -> f2.1
    [EvaluationObject 2][Objective 2][x] -> f2.2
    [EvaluationObject 2][Constraint 1][x] -> g2.1

    [None][Evaluator 1][x] -> IntermediateResults 1
    [Objective 3][x] -> f3
    [Constraint 2][x] -> g2

    See Also
    --------
    CADETProcess.optimization.OptimizationProblem.add_evaluator
    """

    def __init__(
        self,
        use_diskcache: bool = False,
        directory: Optional[str] = None,
        n_shards: int = 1,
    ) -> None:
        """Initialize ResultsCache Object."""
        self.use_diskcache = use_diskcache
        self.directory = directory
        self.n_shards = n_shards
        self.init_cache()

        self.tags = defaultdict(list)

    def init_cache(self) -> None:
        """Initialize ResultsCache."""
        if self.use_diskcache:
            if self.directory is None:
                self.directory = Path(tempfile.mkdtemp(prefix="diskcache-"))

            if self.directory.exists():
                shutil.rmtree(self.directory, ignore_errors=True)

            self.directory.mkdir(exist_ok=True, parents=True)

            if self.n_shards == 1:
                self.cache = Cache(
                    self.directory.as_posix(),
                    disk=DillDisk,
                    disk_min_file_size=2**18,  # 256 kb
                    size_limit=2**36,  # 64 GB
                    tag_index=True,
                )
            else:
                self.cache = FanoutCache(
                    self.directory.as_posix(),
                    shards=self.n_shards,
                    disk=DillDisk,
                    disk_min_file_size=2**18,  # 256 kb
                    size_limit=2**36,  # 64 GB
                    tag_index=True,
                )
            self.directory = self.cache.directory
        else:
            self.cache = {}

    def set(
        self,
        key: Any,
        value: Any,
        tag: Optional[str] = None,
        close: bool = True,
    ) -> None:
        """
        Add entry to cache.

        Parameters
        ----------
        key : hashable
            The key to retrieve the results for.
        value : object
            The value corresponding to the key.
        tag : str, optional
            Tag to associate with result. The default is None.
        close : bool, optional
            If True, database will be closed after operation. The default is True.
        """
        if self.use_diskcache:
            self.cache.set(key, value, tag=tag, expire=None)
        else:
            if tag is not None:
                self.tags[tag].append(key)
            self.cache[key] = value

        if close:
            self.close()

    def get(self, key: Any, close: bool = True) -> Any:
        """
        Get entry from cache.

        Parameters
        ----------
        key : hashable
            The key to retrieve the results for.
        close : bool, optional
            If True, database will be closed after operation. The default is True.

        Returns
        -------
        value : object
            The value corresponding to the key.
        """
        value = self.cache[key]

        if close:
            self.close()

        return value

    def delete(self, key: Any, close: Optional[bool] = True) -> None:
        """
        Remove entry from cache.

        Parameters
        ----------
        key : hashable
            The key to retrieve the results for.
        close : bool, optional
            If True, database will be closed after operation. The default is True.
        """
        if self.use_diskcache:
            found = self.cache.delete(key)
            if not found:
                raise KeyError(key)
        else:
            self.cache.pop(key)

        if close:
            self.close()

    def prune(self, tag: str, close: Optional[bool] = True) -> None:
        """
        Remove tagged entries from cache.

        Parameters
        ----------
        tag : str
            Tag to be removed.
        close : bool, optional
            If True, database will be closed after operation. The default is True.
        """
        if self.use_diskcache:
            self.cache.evict(tag)
        else:
            keys = self.tags.pop(tag, [])

            for key in keys:
                try:
                    self.delete(key, close=False)
                except KeyError:
                    pass

        if close:
            self.close()

    def close(self) -> None:
        """Close cache."""
        if self.use_diskcache:
            self.cache.close()

    def delete_database(self, reinit: Optional[bool] = False) -> None:
        """
        Delte database.

        Parameters
        ----------
        reinit : bool, optional
            If True, reinitialize cache. The default is False.
        """
        self.close()
        if self.use_diskcache:
            try:
                shutil.rmtree(self.directory, ignore_errors=True)
            except FileNotFoundError:
                pass

        if reinit:
            self.init_cache()



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# (batch_elution_optimization_multi)=
# # Optimize Batch Elution Process (Multi-Objective)
#
# ## Setup Optimization Problem

# %%
from CADETProcess.optimization import OptimizationProblem
optimization_problem = OptimizationProblem('batch_elution_multi')

from examples.batch_elution.process import process
optimization_problem.add_evaluation_object(process)

optimization_problem.add_variable('cycle_time', lb=10, ub=600)
optimization_problem.add_variable('feed_duration.time', lb=10, ub=300)

optimization_problem.add_linear_constraint(
    ['feed_duration.time', 'cycle_time'], [1, -1]
)

# %% [markdown]
# ## Setup Simulator

# %%
from CADETProcess.simulator import Cadet
process_simulator = Cadet()
process_simulator.evaluate_stationarity = True

optimization_problem.add_evaluator(process_simulator)

# %% [markdown]
# ## Setup Fractionator

# %%
from CADETProcess.fractionation import FractionationOptimizer
frac_opt = FractionationOptimizer()

optimization_problem.add_evaluator(
    frac_opt,
    kwargs={
        'purity_required': [0.95, 0.95],
        'ignore_failed': False,
        'allow_empty_fractions': False,
    }
)

# %% [markdown]
# ## Setup Objectives

# %%
from CADETProcess.performance import Productivity, Recovery, EluentConsumption

productivity = Productivity()
optimization_problem.add_objective(
    productivity,
    n_objectives=2,
    requires=[process_simulator, frac_opt],
    minimize=False,
)

recovery = Recovery()
optimization_problem.add_objective(
    recovery,
    n_objectives=2,
    requires=[process_simulator, frac_opt],
    minimize=False,
)

eluent_consumption = EluentConsumption()
optimization_problem.add_objective(
    eluent_consumption,
    n_objectives=2,
    requires=[process_simulator, frac_opt],
    minimize=False,
)


# %% [markdown]
# ## Add callback for post-processing

# %%
def callback(fractionation, individual, evaluation_object, callbacks_dir):
    fractionation.plot_fraction_signal(
        file_name=f'{callbacks_dir}/{individual.id}_{evaluation_object}_fractionation.png',
        show=False
    )


optimization_problem.add_callback(
    callback, requires=[process_simulator, frac_opt]
)

# %% [markdown]
# ## Configure Optimizer

# %%
from CADETProcess.optimization import U_NSGA3
optimizer = U_NSGA3()

# %% [markdown]
# ## Run Optimization
#
# ```{note}
# For performance reasons, the optimization is currently not run when building the documentation.
# In future, we will try to sideload pre-computed results to also discuss them here.
# ```
#
# ```
# if __name__ == '__main__':
#     results = optimizer.optimize(
#         optimization_problem,
#         use_checkpoint=True,
#     )
# ```



import hashlib
from typing import Optional

import numpy as np
import numpy.typing as npt
from addict import Dict

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import Bool, Float, Structure, Vector


def hash_array(array: np.ndarray) -> str:
    """
    Compute a hash value for an array of floats using the sha256 hash function.

    Parameters
    ----------
    array : numpy.ndarray
        An array of floats.

    Returns
    -------
    str
        A hash value as a string of hexadecimal characters.

    Examples
    --------
    >>> import numpy as np
    >>> hash_array(np.array([1, 2.0]))
    '3dfc9d56e04dcd01590f48b1b57c9ed9fecb1e94e11d3c3f13cf0fd97b7a9f0f'
    """
    array = np.asarray(array)
    return hashlib.sha256(array.tobytes()).hexdigest()


class Individual(Structure):
    """
    Set of variables evaluated during Optimization.

    Attributes
    ----------
    id : str
        UUID for individual.
    x : np.ndarray
        Variable values in untransformed space.
    x_transformed : np.ndarray
        Independent variable values in transformed space.
    cv_bounds : np.ndarray
        Vound constraint violations.
    cv_lincon : np.ndarray
        Linear constraint violations.
    cv_lineqcon : np.ndarray
        Linear equality constraint violations.
    f : np.ndarray
        Objective values.
    f_min : np.ndarray
        Minimized objective values.
    g : np.ndarray
        Nonlinear constraint values.
    cv_nonlincon : np.ndarray
        Nonlinear constraints violation.
    m : np.ndarray
        Meta score values.
    m_min : np.ndarray
        Minimized meta score values.
    is_feasible : bool
        True, if individual fulfills all constraints.

    See Also
    --------
    CADETProcess.optimization.Population
    """

    x = Vector()
    x_transformed = Vector()
    cv_bounds = Vector()
    cv_lincon = Vector()
    cv_lineqcon = Vector()
    f = Vector()
    f_min = Vector()
    g = Vector()
    cv_nonlincon = Vector()
    cv_nonlincon_tol = Float()
    m = Vector()
    m_min = Vector()
    is_feasible = Bool()

    def __init__(
        self,
        x: npt.ArrayLike,
        f: Optional[npt.ArrayLike] = None,
        g: Optional[npt.ArrayLike] = None,
        f_min: Optional[npt.ArrayLike] = None,
        x_transformed: Optional[npt.ArrayLike] = None,
        cv_bounds: Optional[npt.ArrayLike] = None,
        cv_lincon: Optional[npt.ArrayLike] = None,
        cv_lineqcon: Optional[npt.ArrayLike] = None,
        cv_nonlincon: Optional[npt.ArrayLike] = None,
        m: Optional[npt.ArrayLike] = None,
        m_min: Optional[npt.ArrayLike] = None,
        independent_variable_names: list[str] = None,
        objective_labels: list[str] = None,
        nonlinear_constraint_labels: list[str] = None,
        meta_score_labels: list[str] = None,
        variable_names: list[str] = None,
        is_feasible: bool = True,
    ) -> None:
        """Initialize Individual Object."""
        self.x = x
        if x_transformed is None:
            x_transformed = x
            independent_variable_names = variable_names
        self.x_transformed = x_transformed

        self.cv_bounds = cv_bounds
        self.cv_lincon = cv_lincon
        self.cv_lineqcon = cv_lineqcon

        self.f = f
        if f_min is None:
            f_min = f
        self.f_min = f_min

        self.g = g
        if g is not None and cv_nonlincon is None:
            cv_nonlincon = g
        self.cv_nonlincon = cv_nonlincon

        self.m = m
        if m_min is None:
            m_min = m
        self.m_min = m_min

        if isinstance(variable_names, np.ndarray):
            variable_names = [s.decode() for s in variable_names]
        self.variable_names = variable_names
        if isinstance(independent_variable_names, np.ndarray):
            independent_variable_names = [
                s.decode() for s in independent_variable_names
            ]
        self.independent_variable_names = independent_variable_names
        if isinstance(objective_labels, np.ndarray):
            objective_labels = [s.decode() for s in objective_labels]
        self.objective_labels = objective_labels
        if isinstance(nonlinear_constraint_labels, np.ndarray):
            nonlinear_constraint_labels = [
                s.decode() for s in nonlinear_constraint_labels
            ]
        self.nonlinear_constraint_labels = nonlinear_constraint_labels
        if isinstance(meta_score_labels, np.ndarray):
            meta_score_labels = [s.decode() for s in meta_score_labels]
        self.meta_score_labels = meta_score_labels

        self.id = hash_array(self.x)
        self.is_feasible = is_feasible

    @property
    def id_short(self) -> str:
        """str: Id shortened to the first seven digits."""
        return self.id[0:7]

    @property
    def is_evaluated(self) -> bool:
        """bool: Return True if individual has been evaluated. False otherwise."""
        if self.f is None:
            return False
        else:
            return True

    @property
    def n_x(self) -> int:
        """int: Number of variables."""
        return len(self.x)

    @property
    def n_f(self) -> int:
        """int: Number of objectives."""
        if self.f is None:
            return 0
        return len(self.f)

    @property
    def n_g(self) -> int:
        """int: Number of nonlinear constraints."""
        if self.g is None:
            return 0
        else:
            return len(self.g)

    @property
    def n_m(self) -> int:
        """int: Number of meta scores."""
        if self.m is None:
            return 0
        else:
            return len(self.m)

    @property
    def cv(self) -> np.ndarray:
        """
        All constraint violations combined.

        (cv_bounds, cv_lincon, cv_lineqcon, cv_nonlincon)

        Returns
        -------
        np.ndarray
            All constraint violations combined.
        """
        cvs = (self.cv_bounds, self.cv_lincon, self.cv_lineqcon, self.cv_nonlincon)
        return np.concatenate([cv for cv in cvs if cv is not None])

    @property
    def dimensions(self) -> tuple[int]:
        """tuple: Individual dimensions (n_x, n_f, n_g, n_m)."""
        return (self.n_x, self.n_f, self.n_g, self.n_m)

    @property
    def objectives_minimization_factors(self) -> np.ndarray:
        """np.ndarray: Array indicating objectives transformed to minimization."""
        return self.f_min / self.f

    @property
    def meta_scores_minimization_factors(self) -> np.ndarray:
        """np.ndarray: Array indicating meta sorces transformed to minimization."""
        return self.m_min / self.m

    def dominates(self, other: "Individual") -> bool:
        """
        Determine if individual dominates other.

        Parameters
        ----------
        other : Individual
            Other individual

        Returns
        -------
        dominates : bool
            True if objectives of "self" are not strictly worse than the
            corresponding objectives of "other" and at least one objective is
            strictly better. False otherwise
        """
        if not self.is_evaluated:
            raise CADETProcessError("Individual needs to be evaluated first.")
        if not other.is_evaluated:
            raise CADETProcessError("Other individual needs to be evaluated first.")

        if self.is_feasible and not other.is_feasible:
            return True
        if not self.is_feasible and other.is_feasible:
            return False

        if not self.is_feasible and not other.is_feasible:
            if np.any(self.cv < other.cv):
                better_in_all = np.all(self.cv <= other.cv)
                strictly_better_in_one = np.any(self.cv < other.cv)
                return better_in_all and strictly_better_in_one

        if self.m is not None:
            self_values = self.m_min
            other_values = other.m_min
        else:
            self_values = self.f_min
            other_values = other.f_min

        if np.any(self_values > other_values):
            return False

        if np.any(self_values < other_values):
            return True

        return False

    def is_similar(self, other: "Individual", tol: float = 1e-1) -> bool:
        """
        Determine if individual is similar to other.

        Parameters
        ----------
        other : Individual
            Other individual
        tol : float, optional
            Relative tolerance parameter.
            To reduce number of entries, a rather high rtol is chosen.

        Returns
        -------
        is_similar : bool
            True if individuals are close to each other. False otherwise
        """
        if not tol:
            return False

        similar_x = self.is_similar_x(other, tol)
        similar_f = self.is_similar_f(other, tol)

        if self.g is not None:
            similar_g = self.is_similar_g(other, tol)
        else:
            similar_g = True

        if self.m is not None:
            similar_m = self.is_similar_m(other, tol)
        else:
            similar_m = True

        return similar_x and similar_f and similar_g and similar_m

    def is_similar_x(
        self,
        other: "Individual",
        tol: float = 1e-1,
        use_transformed: bool = False,
    ) -> bool:
        """
        Determine if individual is similar to other based on parameter values.

        Parameters
        ----------
        other : Individual
            Other individual
        tol : float
            Relative tolerance parameter.
            To reduce number of entries, a rather high rtol is chosen.
        use_transformed : bool
            If True, use independent transformed space.
            The default is False.

        Returns
        -------
        is_similar : bool
            True if parameters are close to each other. False otherwise
        """
        similar_x = np.allclose(self.x, other.x, rtol=tol)

        return similar_x

    def is_similar_f(self, other: "Individual", tol: float = 1e-1) -> bool:
        """
        Determine if individual is similar to other based on objective values.

        Parameters
        ----------
        other : Individual
            Other individual
        tol : float
            Relative tolerance parameter.
            To reduce number of entries, a rather high rtol is chosen.

        Returns
        -------
        is_similar : bool
            True if parameters are close to each other. False otherwise
        """
        similar_f = np.allclose(self.f, other.f, rtol=tol)

        return similar_f

    def is_similar_g(self, other: "Individual", tol: float | None = 1e-1) -> bool:
        """
        Determine if individual is similar to other based on constraint values.

        Parameters
        ----------
        other : Individual
            Other individual
        tol : float
            Relative tolerance parameter.
            To reduce number of entries, a rather high rtol is chosen.

        Returns
        -------
        is_similar : bool
            True if parameters are close to each other. False otherwise
        """
        similar_g = np.allclose(self.g, other.g, rtol=tol)

        return similar_g

    def is_similar_m(self, other: "Individual", tol: float | None = 1e-1) -> bool:
        """
        Determine if individual is similar to other based on meta score values.

        Parameters
        ----------
        other : Individual
            Other individual
        tol : float
            Relative tolerance parameter.
            To reduce number of entries, a rather high rtol is chosen.

        Returns
        -------
        is_similar : bool
            True if parameters are close to each other. False otherwise
        """
        similar_m = np.allclose(self.m, other.m, rtol=tol)

        return similar_m

    def __str__(self) -> str:
        """str: String representation of the individual."""
        return str(list(self.x))

    def __repr__(self) -> str:
        """str: String representation of the individual."""
        if self.g is None:
            return f"{self.__class__.__name__}({self.x}, {self.f})"
        else:
            return f"{self.__class__.__name__}({self.x}, {self.f}, {self.g})"

    def to_dict(self) -> dict:
        """
        Convert individual to a dictionary.

        Returns
        -------
        dict: A dictionary representation of the individual's attributes.
        """
        data = Dict()

        data.x = self.x
        data.x_transformed = self.x_transformed

        data.cv_bounds = self.cv_bounds
        data.cv_lincon = self.cv_lincon
        data.cv_lineqcon = self.cv_lineqcon

        data.f = self.f
        data.f_min = self.f_min

        if self.g is not None:
            data.g = self.g
            data.cv_nonlincon = self.cv_nonlincon

        if self.m is not None:
            data.m = self.m
            data.m_min = self.m_min

        data.variable_names = self.variable_names
        data.independent_variable_names = self.independent_variable_names
        if self.objective_labels is not None:
            data.objective_labels = self.objective_labels
        if self.nonlinear_constraint_labels is not None:
            data.nonlinear_constraint_labels = self.nonlinear_constraint_labels
        if self.meta_score_labels is not None:
            data.meta_score_labels = self.meta_score_labels

        data.is_feasible = self.is_feasible

        return data

    @classmethod
    def from_dict(cls, data: dict) -> "Individual":
        """
        Create Individual from dictionary representation of its attributes.

        Parameters
        ----------
        data : dict
            A dictionary representation of the individual's attributes.

        Returns
        -------
        individual
            Individual idual created from the dictionary.
        """
        return cls(**data)



"""
================================================================
Discretization (:mod:`CADETProcess.processModel.discretization`)
================================================================

The following module is a collection of classes that define different discretization
schemes for :mod:`UnitOperations <CADETProcess.processModel.unitOperation>` for the
CADETProcess software package.

The :class:`~CADETProcess.processModel.DiscretizationParametersBase` class is the base
class for all other classes in this module and defines some common parameters.
Specific parameters for each scheme are defined as attributes of each class.

"""  # noqa

from CADETProcess.dataStructure import (
    Bool,
    Constant,
    RangedInteger,
    SizedRangedList,
    Structure,
    Switch,
    UnsignedFloat,
    UnsignedInteger,
    frozen_attributes,
)

__all__ = [
    "DiscretizationParametersBase",
    "NoDiscretization",
    "LRMDiscretizationFV",
    "LRMDiscretizationDG",
    "LRMPDiscretizationFV",
    "LRMPDiscretizationDG",
    "GRMDiscretizationFV",
    "GRMDiscretizationDG",
    "WenoParameters",
    "ConsistencySolverParameters",
    "DGMixin",
    "MCTDiscretizationFV",
]


@frozen_attributes
class DiscretizationParametersBase(Structure):
    """
    Base class for storing discretization parameters.

    Attributes
    ----------
    dimensionality : dict
        Dimensionality of the parameters
    weno_parameters : WenoParameters
        Parameters for the WENO scheme.
    consistency_solver: ConsistencySolverParameters
        Consistency solver parameters for Cadet.
    """

    _dimensionality = []

    def __init__(self) -> None:
        """Initialize a new DiscretizationParametersBase instance."""
        self.weno_parameters = WenoParameters()
        self.consistency_solver = ConsistencySolverParameters()

        super().__init__()

    @property
    def dimensionality(self) -> dict:
        """dict: Dimensionality of the parameters."""
        dim = {}
        for d in self._dimensionality:
            v = getattr(self, d)
            if v is None:
                continue
            dim[d] = v

        return dim

    @property
    def parameters(self) -> dict:
        """dict: Dictionary with parameter values."""
        parameters = super().parameters
        parameters["weno"] = self.weno_parameters.parameters
        parameters["consistency_solver"] = self.consistency_solver.parameters

        return parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        try:
            self.weno_parameters.parameters = parameters.pop("weno")
        except KeyError:
            pass
        try:
            self.consistency_solver.parameters = parameters.pop("consistency_solver")
        except KeyError:
            pass

        super(DiscretizationParametersBase, self.__class__).parameters.fset(
            self, parameters
        )


class NoDiscretization(DiscretizationParametersBase):
    """Class for unit operations without spatial discretization."""

    pass


class DGMixin(DiscretizationParametersBase):
    """Mixin DG discretization parameters."""

    pass


class LRMDiscretizationFV(DiscretizationParametersBase):
    """
    Discretization parameters of the FV version of the LRM.

    Attributes
    ----------
    ncol : UnsignedInteger, optional
        Number of axial column discretization cells. Default is 100.
    use_analytic_jacobian : Bool, optional
        If True, use analytically computed Jacobian matrix (faster).
        If False, use Jacobians generated by algorithmic differentiation (slower).
        Default is True.
    reconstruction : Switch, optional
        Method for spatial reconstruction. Valid values are 'WENO' (Weighted
        Essentially Non-Oscillatory). Default is 'WENO'.
    """

    spatial_method = Constant(value="FV")
    ncol = UnsignedInteger(default=100)
    use_analytic_jacobian = Bool(default=True)
    reconstruction = Switch(default="WENO", valid=["WENO"])

    _parameters = DiscretizationParametersBase._parameters + [
        "ncol",
        "use_analytic_jacobian",
        "reconstruction",
    ]
    _dimensionality = ["ncol"]


class LRMDiscretizationDG(DGMixin):
    """
    Discretization parameters of the DG version of the LRM.

    Attributes
    ----------
    nelem : UnsignedInteger, optional
        Number of axial column elements. Default is 16.
    use_analytic_jacobian : Bool, optional
        If True, use analytically computed Jacobian matrix (faster).
        If False, use Jacobians generated by algorithmic differentiation (slower).
        Default is True.
    polydeg : UnsignedInteger, optional
        Degree of the polynomial used for axial discretization. Default is 4.
    exact_integration : Bool, optional
        Whether to use exact integration for the axial discretization.
        Default is False.

    See Also
    --------
    CADETProcess.processModel.LRMPDiscretizationFV
    CADETProcess.processModel.LumpedRateModelWithPores
    """

    spatial_method = Constant(value="DG")
    nelem = UnsignedInteger(default=16)
    use_analytic_jacobian = Bool(default=True)
    polydeg = UnsignedInteger(default=4)
    exact_integration = Bool(default=False)

    _parameters = DiscretizationParametersBase._parameters + [
        "nelem",
        "use_analytic_jacobian",
        "polydeg",
        "exact_integration",
    ]
    _dimensionality = ["axial_dof"]

    @property
    def axial_dof(self) -> int:
        """int: Number of degrees of freedom in the axial discretization."""
        return self.nelem * (self.polydeg + 1)


class LRMPDiscretizationFV(DiscretizationParametersBase):
    """
    Discretization parameters of the FV version of the LRMP.

    Attributes
    ----------
    ncol : UnsignedInteger, optional
        Number of axial column discretization cells. Default is 100.
    par_geom : Switch, optional
        The geometry of the particles in the model.
        Valid values are 'SPHERE', 'CYLINDER', and 'SLAB'.
        Default is 'SPHERE'.
    use_analytic_jacobian : Bool, optional
        If True, use analytically computed Jacobian matrix (faster).
        If False, use Jacobians generated by algorithmic differentiation (slower).
        Default is True.
    reconstruction : Switch, optional
        Method for spatial reconstruction. Valid values are 'WENO' (Weighted
        Essentially Non-Oscillatory). Default is 'WENO'.
    gs_type : Bool, optional
        Type of Gram-Schmidt orthogonalization.
        If 0, use classical Gram-Schmidt.
        If 1, use modified Gram-Schmidt.
        The default is 1.
    max_krylov : UnsignedInteger, optional
        Size of the Krylov subspace in the iterative linear GMRES solver.
        If 0, max_krylov = NCOL * NCOMP * NPARTYPE is used.
        The default is 0.
    max_restarts : UnsignedInteger, optional
        Maximum number of restarts to use for the GMRES method. Default is 10.
    schur_safety : UnsignedFloat, optional
        Safety factor for the Schur complement solver. Default is 1.0e-8.

    See Also
    --------
    CADETProcess.processModel.LRMPDiscretizationDG
    CADETProcess.processModel.LumpedRateModelWithPores
    """

    spatial_method = Constant(value="FV")
    ncol = UnsignedInteger(default=100)

    par_geom = Switch(default="SPHERE", valid=["SPHERE", "CYLINDER", "SLAB"])

    use_analytic_jacobian = Bool(default=True)
    reconstruction = Switch(default="WENO", valid=["WENO"])

    gs_type = Bool(default=True)
    max_krylov = UnsignedInteger(default=0)
    max_restarts = UnsignedInteger(default=10)
    schur_safety = UnsignedFloat(default=1.0e-8)

    _parameters = DiscretizationParametersBase._parameters + [
        "ncol",
        "par_geom",
        "use_analytic_jacobian",
        "reconstruction",
        "gs_type",
        "max_krylov",
        "max_restarts",
        "schur_safety",
    ]
    _dimensionality = ["ncol"]


class LRMPDiscretizationDG(DGMixin):
    """
    Discretization parameters of the DG version of the LRMP.

    Attributes
    ----------
    nelem : UnsignedInteger, optional
        Number of axial column elements. Default is 16.
    par_geom : Switch, optional
        The geometry of the particles in the model.
        Valid values are 'SPHERE', 'CYLINDER', and 'SLAB'.
        Default is 'SPHERE'.
    use_analytic_jacobian : Bool, optional
        If True, use analytically computed Jacobian matrix (faster).
        If False, use Jacobians generated by algorithmic differentiation (slower).
        Default is True.
    polydeg : UnsignedInteger, optional
        Degree of the polynomial used for spatial discretization. Default is 4.
    exact_integration : Bool, optional
        Whether to use exact integration for the spatial discretization.
        Default is False.

    See Also
    --------
    CADETProcess.processModel.LRMPDiscretizationFV
    CADETProcess.processModel.LumpedRateModelWithPores
    """

    spatial_method = Constant(value="DG")
    nelem = UnsignedInteger(default=16)

    par_geom = Switch(default="SPHERE", valid=["SPHERE", "CYLINDER", "SLAB"])

    use_analytic_jacobian = Bool(default=True)
    polydeg = UnsignedInteger(default=4)
    exact_integration = Bool(default=False)

    _parameters = DiscretizationParametersBase._parameters + [
        "nelem",
        "par_geom",
        "use_analytic_jacobian",
        "polydeg",
        "exact_integration",
    ]
    _dimensionality = ["axial_dof"]

    @property
    def axial_dof(self) -> int:
        """int: Number of axial degrees of freedom in the spatial discretization."""
        return self.nelem * (self.polydeg + 1)


class GRMDiscretizationFV(DiscretizationParametersBase):
    """
    Discretization parameters of the FV version of the LRMP.

    Attributes
    ----------
    ncol : UnsignedInteger, optional
        Number of axial column discretization cells. Default is 100.
    npar : UnsignedInteger, optional
        Number of discretization cells in the radial direction. Default is 5.
    par_geom : Switch, optional
        The geometry of the particles in the model.
        Valid values are 'SPHERE', 'CYLINDER', and 'SLAB'.
        Default is 'SPHERE'.
    par_disc_type : Switch, optional
        Discretization scheme inside the particles for all or each particle type.
        Valid values are 'EQUIDISTANT_PAR', 'EQUIVOLUME_PAR', and 'USER_DEFINED_PAR'.
        Default is 'EQUIDISTANT_PAR'.
    par_disc_vector : SizedRangedList, optional
        Node coordinates for the cell boundaries
        (ignored if `par_disc_type` != `USER_DEFINED_PAR).
        The coordinates are relative and have to include the endpoints `0` and `1`.
        They are later linearly mapped to the true radial range.
        The coordinates for each particle type are appended to one long vector in
        type-major ordering.
        Default is a uniformly spaced vector with `npar+1` points between 0 and 1.
    par_boundary_order : RangedInteger, optional
        The order of the boundary scheme used to discretize the particles.
        Valid values are 1 (first order) and 2 (second order).
        Default is 2.
    use_analytic_jacobian : Bool, optional
        If True, use analytically computed Jacobian matrix (faster).
        If False, use Jacobians generated by algorithmic differentiation (slower).
        Default is True.
    reconstruction : Switch, optional
        Method for spatial reconstruction. Valid values are 'WENO' (Weighted
        Essentially Non-Oscillatory). Default is 'WENO'.
    gs_type : Bool, optional
        Type of Gram-Schmidt orthogonalization.
        If 0, use classical Gram-Schmidt.
        If 1, use modified Gram-Schmidt.
        The default is 1.
    max_krylov : UnsignedInteger, optional
        Size of the Krylov subspace in the iterative linear GMRES solver.
        If 0, max_krylov = ncol * ncomp * npar is used, where ncomp is the
        number of components in the model.
        Default is 0.
    max_restarts : UnsignedInteger, optional
        Maximum number of restarts to use for the GMRES method. Default is 10.
    schur_safety : UnsignedFloat, optional
        Safety factor for the Schur complement solver. Default is 1.0e-8.
    fix_zero_surface_diffusion : Bool, optional
        If True, fix the surface diffusion coefficient of particles with zero
        surface diffusion to a small positive value. Default is False.

    See Also
    --------
    CADETProcess.processModel.LRMPDiscretizationDG
    CADETProcess.processModel.LumpedRateModelWithPores
    """

    spatial_method = Constant(value="FV")
    ncol = UnsignedInteger(default=100)
    npar = UnsignedInteger(default=5)

    par_geom = Switch(default="SPHERE", valid=["SPHERE", "CYLINDER", "SLAB"])
    par_disc_type = Switch(
        default="EQUIDISTANT_PAR",
        valid=["EQUIDISTANT_PAR", "EQUIVOLUME_PAR", "USER_DEFINED_PAR"],
    )
    par_disc_vector = SizedRangedList(
        lb=0, ub=1, size="par_disc_vector_length", is_optional=True
    )

    par_boundary_order = RangedInteger(lb=1, ub=2, default=2)

    use_analytic_jacobian = Bool(default=True)
    reconstruction = Switch(default="WENO", valid=["WENO"])

    gs_type = Bool(default=True)
    max_krylov = UnsignedInteger(default=0)
    max_restarts = UnsignedInteger(default=10)
    schur_safety = UnsignedFloat(default=1.0e-8)

    fix_zero_surface_diffusion = Bool(default=False)

    _parameters = DiscretizationParametersBase._parameters + [
        "ncol",
        "npar",
        "par_geom",
        "par_disc_type",
        "par_disc_vector",
        "par_boundary_order",
        "use_analytic_jacobian",
        "reconstruction",
        "gs_type",
        "max_krylov",
        "max_restarts",
        "schur_safety",
        "fix_zero_surface_diffusion",
    ]
    _required_parameters = [
        "ncol",
        "npar",
        "par_geom",
        "par_disc_type",
        "par_boundary_order",
        "use_analytic_jacobian",
        "reconstruction",
        "gs_type",
        "max_krylov",
        "max_restarts",
        "schur_safety",
        "fix_zero_surface_diffusion",
    ]
    _dimensionality = ["ncol", "npar"]

    @property
    def par_disc_vector_length(self) -> int:
        """int: Number of entries in the particle discretization vector."""
        return self.npar + 1


class GRMDiscretizationDG(DGMixin):
    """
    Discretization parameters of the DG version of the GRM.

    Attributes
    ----------
    nelem : UnsignedInteger, optional
        Number of axial column elements. Default is 16.
    par_nelem : UnsignedInteger, optional
        Number of particle (radial) discretization cells for each particle type.
        Default is 1.
    par_geom : Switch, optional
        The geometry of the particles in the model.
        Valid values are 'SPHERE', 'CYLINDER', and 'SLAB'.
        Default is 'SPHERE'.
    use_analytic_jacobian : Bool, optional
        If True, use analytically computed Jacobian matrix (faster).
        If False, use Jacobians generated by algorithmic differentiation (slower).
        Default is True.
    polydeg : UnsignedInteger, optional
        Degree of the polynomial used for axial discretization. Default is 4.
    par_polydeg : UnsignedInteger, optional
        Degree of the polynomial used for particle radial discretization. Default is 4.
    exact_integration : Bool, optional
        Whether to use exact integration for the axial discretization.
        Default is False.
    par_exact_integration : Bool, optional
        Whether to use exact integration for the particle radial discretization.
        Default is True.
    fix_zero_surface_diffusion : Bool, optional
        Whether to fix zero surface diffusion for particles. Default is False.
        If True, the parameters must not become non-zero during this or subsequent
        simulation runs. The internal data structures are optimized for a more efficient
        simulation.
        Default is False (optimization disabled in favor of flexibility).

    See Also
    --------
    CADETProcess.processModel.GRMDiscretizationDG
    CADETProcess.processModel.GeneralRateModel
    """

    spatial_method = Constant(value="DG")
    nelem = UnsignedInteger(default=16)
    par_nelem = UnsignedInteger(default=1)

    par_geom = Switch(default="SPHERE", valid=["SPHERE", "CYLINDER", "SLAB"])
    par_disc_type = Switch(
        default="EQUIDISTANT_PAR",
        valid=["EQUIDISTANT_PAR", "EQUIVOLUME_PAR", "USER_DEFINED_PAR"],
    )
    par_disc_vector = SizedRangedList(lb=0, ub=1, size="par_disc_vector_length")

    par_boundary_order = RangedInteger(lb=1, ub=2, default=2)

    use_analytic_jacobian = Bool(default=True)
    polydeg = UnsignedInteger(default=4)
    par_polydeg = UnsignedInteger(default=4)
    exact_integration = Bool(default=False)
    par_exact_integration = Bool(default=True)

    fix_zero_surface_diffusion = Bool(default=False)

    _parameters = DiscretizationParametersBase._parameters + [
        "nelem",
        "par_nelem",
        "par_geom",
        "par_disc_type",
        "par_disc_vector",
        "par_boundary_order",
        "use_analytic_jacobian",
        "polydeg",
        "par_polydeg",
        "exact_integration",
        "par_exact_integration",
        "fix_zero_surface_diffusion",
    ]
    _dimensionality = ["axial_dof", "par_dof"]

    @property
    def axial_dof(self) -> int:
        """int: Number of axial degrees of freedom in the axial discretization."""
        return self.nelem * (self.polydeg + 1)

    @property
    def par_dof(self) -> int:
        """int: Number of particle degrees of freedom in the axial discretization."""
        return self.axial_dof * self.par_disc_vector_length

    @property
    def par_disc_vector_length(self) -> int:
        """int: Number of entries in the particle discretization vector."""
        return self.par_nelem * (self.par_polydeg + 1)


@frozen_attributes
class WenoParameters(Structure):
    """
    Discretization parameters for the WENO scheme.

    Attributes
    ----------
    boundary_model : Switch, optional
        Specifies the method for dealing with boundary cells.
        Valid values are:
        0: Lower WENO order (stable)
        1: Zero weights (unstable for small `D_ax`)
        2: Zero weights for p != 0 (stable?)
        3: Large ghost points
        Default is 0.
    weno_eps : UnsignedFloat, optional
        A small positive number used to avoid division by zero in the WENO scheme.
        Default is 1e-10.
    weno_order : UnsignedInteger, optional
        Order of the WENO scheme. Valid values are:
        1: Standard upwind scheme (order 1)
        2: WENO 2 (order 3)
        3: WENO 3 (order 5)
        Default is 3.

    See Also
    --------
    Structure
    """

    boundary_model = UnsignedInteger(default=0, ub=3)
    weno_eps = UnsignedFloat(default=1e-10)
    weno_order = UnsignedInteger(default=3, ub=3)
    _parameters = ["boundary_model", "weno_eps", "weno_order"]


@frozen_attributes
class ConsistencySolverParameters(Structure):
    """
    A class for defining the consistency solver parameters for Cadet.

    Parameters
    ----------
    solver_name : Switch, optional
        Name of the solver.
        Valid values are 'LEVMAR', 'ATRN_RES', 'ATRN_ERR', and 'COMPOSITE'.
        The default is 'LEVMAR'
    init_damping : UnsignedFloat, optional
        The initial damping parameter. Default is 0.01.
    min_damping : UnsignedFloat, optional
        The minimum damping parameter. Default is 0.0001.
    max_iterations : UnsignedFloat, optional
        The maximum number of iterations. Default is 50.
    subsolvers : Switch, optional
        Vector with names of solvers for the composite solver
        (only required for composite solver).
        Valid values are 'LEVMAR', 'ATRN_RES', 'ATRN_ERR', and 'COMPOSITE'.
        The default is 'LEVMAR'

    See Also
    --------
    Structure
    """

    solver_name = Switch(
        default="LEVMAR", valid=["LEVMAR", "ATRN_RES", "ARTN_ERR", "COMPOSITE"]
    )
    init_damping = UnsignedFloat(default=0.01)
    min_damping = UnsignedFloat(default=0.0001)
    max_iterations = UnsignedInteger(default=50)
    subsolvers = Switch(default="LEVMAR", valid=["LEVMAR", "ATRN_RES", "ARTN_ERR"])

    _parameters = [
        "solver_name",
        "init_damping",
        "min_damping",
        "max_iterations",
        "subsolvers",
    ]


class MCTDiscretizationFV(DiscretizationParametersBase):
    """
    Discretization parameters of the FV version of the MCT.

    Attributes
    ----------
    ncol : UnsignedInteger, optional
        Number of axial column discretization cells. Default is 100.
    use_analytic_jacobian : Bool, optional
        If True, use analytically computed Jacobian matrix (faster).
        If False, use Jacobians generated by algorithmic differentiation (slower).
        Default is True.
    reconstruction : Switch, optional
        Method for spatial reconstruction. Valid values are 'WENO' (Weighted
        Essentially Non-Oscillatory). Default is 'WENO'.
    """

    spatial_method = Constant(value="FV")
    ncol = UnsignedInteger(default=100)
    use_analytic_jacobian = Bool(default=True)
    reconstruction = Switch(default="WENO", valid=["WENO"])

    _parameters = ["ncol", "use_analytic_jacobian", "reconstruction"]
    _dimensionality = ["ncol"]



import copy

import numpy as np
import scipy.signal

from CADETProcess.solution import SolutionIO


def find_peaks(
    solution: SolutionIO,
    normalize: bool = True,
    prominence: float = 0.5,
    find_minima: bool = False,
) -> list[list[tuple[float, float]]]:
    """
    Find peaks in solution.

    Parameters
    ----------
    solution : SolutionIO
        Solution object.
    normalize : bool, optional
        If true, normalize data to maximum value (for each component).
        The default is True.
    prominence : float, optional
        Required prominence to  detekt peak. The default is 0.5.
    find_minima : bool, optional
        Find negative peaks/minima of solution. The default is False.

    Returns
    -------
    peaks : list
        List with list of (time, height) for each peak for every component.
        Regardless of normalization, the actual peak height is returned.
    """
    solution_original = solution
    solution = copy.deepcopy(solution)

    if normalize:
        solution = solution.normalize()

    peaks = []
    for i in range(solution.component_system.n_comp):
        sol = solution.solution[:, i].copy()

        if find_minima:
            sol *= -1

        peak_indices, _ = scipy.signal.find_peaks(sol, prominence=prominence)
        if len(peak_indices) == 0:
            peak_indices = [np.argmax(sol)]
        time = solution.time[peak_indices]
        peak_heights = solution_original.solution[peak_indices, i]

        peaks.append([(t, h) for t, h in zip(time, peak_heights)])

    return peaks


def find_breakthroughs(
    solution: SolutionIO,
    normalize: bool = True,
    threshold: float = 0.95,
) -> list[(float, float)]:
    """
    Find breakthroughs in solution.

    Parameters
    ----------
    solution : SolutionIO
        Solution object.
    normalize : bool, optional
        If true, normalize data to maximum value (for each component).
        The default is True.
    threshold : float, optional
        Percentage of maximum concentration that needs to be reached to be
        considered as breakthrough. The default is 0.95.

    Returns
    -------
    breakthrough : list
        List with (time, height) for breakthrough of every component.
        Regardless of normalization, the actual breakthroug height is returned.
    """
    solution_original = solution
    solution = copy.deepcopy(solution)

    if normalize:
        solution = solution.normalize()

    breakthrough = []
    for i in range(solution.component_system.n_comp):
        sol = solution.solution[:, i].copy()

        breakthrough_indices = np.where(sol > threshold * np.max(sol))[0][0]
        if len(breakthrough_indices) == 0:
            breakthrough_indices = [np.argmax(sol)]
        time = solution.time[breakthrough_indices]
        breakthrough_height = solution_original.solution[breakthrough_indices, i]

        breakthrough.append((time, breakthrough_height))

    return breakthrough
